import uuid
import csv
import pickle
import logging
import hashlib
import requests
import datetime
import pathlib
import json

from anytree import Node
from logging.handlers import RotatingFileHandler
from chardet.universaldetector import UniversalDetector

INFO_LEVEL = 20
LOG_FILE = 'udvalg.log'
BASE_URL = 'http://localhost:5000/service/'
CACHE = {}
SESSION = None 

logger = logging.getLogger()
log_format = logging.Formatter(
    '%(asctime)s | %(name)s |  %(levelname)s: %(message)s'
)
logger.setLevel(logging.DEBUG)
activity_log_handler = RotatingFileHandler(
    filename=LOG_FILE,
    maxBytes=1000000
)
activity_log_handler.setFormatter(log_format)
activity_log_handler.setLevel(INFO_LEVEL)
logger.addHandler(activity_log_handler)




def _find_class(find_facet, find_class):
    if find_class in CACHE:
        return CACHE[find_class]
    uuid = None
    url = BASE_URL + 'o/{}/f/{}'
    response = SESSION.get(url.format(ROOT, find_facet))
    response.raise_for_status()
    response = response.json()
    for actual_class in response['data']['items']:
        if actual_class['name'] == find_class:
            uuid = actual_class['uuid']
            CACHE[find_class] = uuid
    return uuid


def _mo_lookup(uuid, details=''):
    if not details:
        url = BASE_URL + 'e/{}'
    else:
        url = BASE_URL + 'e/{}/details/' + details
    response = SESSION.get(url.format(uuid))
    response.raise_for_status()
    return(response.json())


def _find_org():
    url = BASE_URL + 'o'
    response = SESSION.get(url)
    response.raise_for_status()
    response = response.json()
    assert(len(response) == 1)
    uuid = response[0]['uuid']
    return(uuid)


def _search_mo_name(name, user_key):
    url = BASE_URL + 'o/{}/e?query={}'
    response = SESSION.get(url.format(ROOT, name))
    result = response.json()
    if len(result['items']) == 1:
        return result['items'][0]['uuid']
    # Did not succeed with simple search, try user_Key
    response = SESSION.get(url.format(ROOT, user_key))
    response.raise_for_status()
    result = response.json()
    for employee in result['items']:
        uuid = employee['uuid']
        mo_user = _mo_lookup(uuid)
        if mo_user['user_key'] == user_key:
            return(employee['uuid'])
    # Still no success, give up and return None
    return None


def _load_csv(file_name):
    rows = []
    detector = UniversalDetector()
    with open(file_name, 'rb') as csvfile:
        for row in csvfile:
            detector.feed(row)
            if detector.done:
                break
    detector.close()
    encoding = detector.result['encoding']

    with open(file_name, encoding=encoding) as csvfile:
        reader = csv.DictReader(csvfile, delimiter=';')
        for row in reader:
            rows.append(row)
    return rows


def generate_uuid(value, org_name):
    """
    Code almost identical to this also lives in the Opus importer.
    """
    base_hash = hashlib.md5(org_name.encode())

    base_digest = base_hash.hexdigest()
    base_uuid = uuid.UUID(base_digest)

    combined_value = (str(base_uuid) + str(value)).encode()
    value_hash = hashlib.md5(combined_value)
    value_digest = value_hash.hexdigest()
    value_uuid = str(uuid.UUID(value_digest))
    return value_uuid


def _create_mo_ou(name, parent, org_type, bvn):
    uuid = generate_uuid(bvn, ROOT)
    ou_type = _find_class(find_facet='org_unit_type', find_class=org_type)
    if parent == 'root':
        parent = ROOT
    payload = {
        'uuid': uuid,
        'user_key': str(bvn),
        'name': '{} {}'.format(org_type, name),
        'org_unit_type': {'uuid': ou_type},
        'parent': {'uuid': parent},
        'validity': {'from': '1930-01-01',
                     'to':  None}
    }

    url = BASE_URL + 'ou/create'
    params = {'force': 1}
    response = SESSION.post(url, json=payload, params=params)
    response.raise_for_status()
    uuid = response.json()
    return uuid


def _create_mo_association(user, org_unit, association_type, from_string):
    response = _mo_lookup(user, details='engagement')
    if response:
        job_function = response[0]['job_function']['uuid']
        payload = [
            {
                'type': 'association',
                'org_unit': {'uuid': org_unit},
                'person': {'uuid': user},
                'association_type': {'uuid': association_type},
                'job_function': {'uuid': job_function},
                'validity': {
                    'from': from_string,
                    'to': None
                }
            }
        ]
        url = BASE_URL + 'details/create'
        params = {'force': 1}
        response = SESSION.post(url, json=payload, params=params)
        response.raise_for_status()
        uuid = response.json()
        return uuid
    else:
        logger.warning('User {} has no engagements'.format(user))
        return None


def _create_mo_role(user, org_unit, role_type, from_string):
    response = _mo_lookup(user, details='engagement')
    if response:
        job_function = response[0]['job_function']['uuid']
        payload = [
            {
                'type': 'role',
                'org_unit': {'uuid': org_unit},
                'person': {'uuid': user},
                'role_type': {'uuid': role_type},
                'job_function': {'uuid': job_function},
                'validity': {
                    'from': from_string,
                    'to': None
                }
            }
        ]
        url = BASE_URL + 'details/create'
        params = {'force': 1}
        response = SESSION.post(url, json=payload, params=params)
        response.raise_for_status()
        uuid = response.json()
        return uuid
    else:
        logger.warning('User {} has no engagements'.format(user))
        return None


def create_udvalg(nodes, file_name):
    rows = _load_csv(file_name)
    for row in rows:
        if ('Formand' in row) and (row['Formand'] == '1'):
            association_type = _find_class('association_type', 'Formand')
        elif ('Næstformand' in row) and (row['Næstformand'] == '1'):
            association_type = _find_class('association_type', 'Næstformand')
        else:
            association_type = _find_class('association_type', 'Medlem')

        if ('TR' in row) and (row['TR'] == '1'):
            role_type = _find_class('role_type', 'Tillidrepræsentant')
        else:
            role_type = None

        org_id = int(row['Id'])
        uuid = _search_mo_name(row['Fornavn'] + ' ' + row['Efternavn'],
                               row['BrugerID'])
        try:
            from_string = datetime.datetime.strftime(
                datetime.datetime.strptime(row['StartDato'], '%d-%b-%y'),
                '%Y-%m-%d'
            )
        except ValueError:
            from_string = '1930-01-01'

        if uuid:
            nodes[uuid] = Node(row['Fornavn'] + ' ' + row['Efternavn'],
                               uuid=uuid,
                               org_type=row['OrgType'],
                               parent=nodes[org_id])
            _create_mo_association(uuid,
                                   nodes[org_id].uuid,
                                   association_type,
                                   from_string)
            if role_type:
                _create_mo_role(uuid, nodes[org_id].uuid, role_type, from_string)

        else:
            logger.warning(
                'Error: {} {}, bvn: {}'.format(row['Fornavn'],
                                               row['Efternavn'],
                                               row['BrugerID'])
            )
    return nodes


def create_tree(file_name):
    nodes = {}
    rows = _load_csv(file_name)
    while rows:
        new = {}
        remaining_nodes = []
        for row in rows:
            org_type = row['OrgType'].strip()
            id_nr = int(row['Id'])
            parent = int(row['ParentID']) if row['ParentID'] else None
            if parent is None:
                uuid = _create_mo_ou(row['OrgEnhed'], parent='root',
                                     org_type=org_type, bvn=id_nr)
                new[id_nr] = Node(row['OrgEnhed'],
                                  uuid=uuid, org_type=org_type)
            elif parent in nodes:
                uuid = _create_mo_ou(row['OrgEnhed'],
                                     parent=nodes[parent].uuid,
                                     org_type=org_type, bvn=id_nr)
                new[id_nr] = Node(row['OrgEnhed'],
                                  uuid=uuid, org_type=org_type,
                                  parent=nodes[parent])
            else:
                remaining_nodes.append(row)
        rows = remaining_nodes
        nodes.update(new)
    return nodes


if __name__ == '__main__':
    logger.info('Program started')

    settingsfile = pathlib.Path("settings") / "settings.json"
    settings = json.loads(settingsfile.read_text())


    SESSION = requests.Session()
    if settings.get("crontab.SAML_TOKEN", None) is not None:
        SESSION.headers["SESSION"] = settings["crontab.SAML_TOKEN"]


    ROOT = _find_org()

    orgtyper_file = '/opt/customer/dataimport/ballerup_udvalg/OrgTyper.csv'
    amr_medlemmer_file = '/opt/customer/dataimport/ballerup_udvalg/AMR-medlemmer.csv'
    med_medlemmer_file = '/opt/customer/dataimport/ballerup_udvalg/MED-medlemmer.csv'

    if True:
        nodes = create_tree(orgtyper_file)
        with open('nodes.p', 'wb') as f:
            pickle.dump(nodes, f, pickle.HIGHEST_PROTOCOL)

    with open('nodes.p', 'rb') as f:
        nodes = pickle.load(f)

    logger.info('Create AMR')
    nodes = create_udvalg(nodes, amr_medlemmer_file)
    logger.info('Create MED')
    nodes = create_udvalg(nodes, med_medlemmer_file)

    # root = min(nodes.keys())
    # from anytree import RenderTree
    # print(RenderTree(nodes[root]))
    logger.info('Program completed.')



import copy
from unittest import mock
from unittest import TestCase

from freezegun import freeze_time
from jinja2.exceptions import UndefinedError
from more_itertools import first_true
from parameterized import parameterized
from ra_utils.lazy_dict import LazyDict

from ..ad_exceptions import CprNotFoundInADException
from ..ad_exceptions import CprNotNotUnique
from ..ad_exceptions import NoPrimaryEngagementException
from ..ad_exceptions import SamAccountNameNotUnique
from ..ad_template_engine import illegal_parameters
from ..ad_writer import ADWriter
from ..ad_writer import LoraCacheSource
from ..user_names import UserNameSetInAD
from ..utils import AttrDict
from .mocks import MockADParameterReader
from .mocks import MockMORESTSource
from .test_utils import dict_modifier
from .test_utils import mo_modifier
from .test_utils import TestADWriterMixin


JOB_TITLE_AD_FIELD_NAME = "titel"
JOB_TITLE_TEMPLATE = "{{ ad_values.get('titel') or mo_values['title'] }}"

_SYNC_TIMESTAMP = "2020-01-01 12:30:00"


class TestADWriter(TestCase, TestADWriterMixin):
    def setUp(self):
        self._setup_adwriter()

    def _verify_identical_common_code(self, num_expected_scripts, num_common_lines=5):
        """Verify that common code in all scripts is identical.

        I.e. that all scripts start with the same num_common_lines lines.
        """
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)
        # 1. Convert each script from a string into a list of strings (lines)
        lines = [script.split("\n") for script in self.ad_writer.scripts]
        self.assertGreaterEqual(len(lines[0]), num_common_lines)
        self.assertEqual(len(lines), num_expected_scripts)
        # 2. Get the common lines (first `num_common_lines` lines in each
        # script.)
        common_lines = [x[:num_common_lines] for x in lines]
        self.assertEqual(len(common_lines[0]), num_common_lines)
        self.assertEqual(len(common_lines), num_expected_scripts)
        # 3. Zip the lines producing `num_common_lines` tuples of 'n' elements,
        # where 'n' is len(scripts).
        zip_lines = list(zip(*common_lines))
        self.assertEqual(len(zip_lines[0]), num_expected_scripts)
        self.assertEqual(len(zip_lines), num_common_lines)
        # Check that all zip_lines are identical
        for zip_line in zip_lines:
            self.assertEqual(len(set(zip_line)), 1)
        # Return common code
        return common_lines[0]

    def _add_to_template_to_ad_fields(self, ad_field_name, template):
        return dict_modifier(
            {"integrations.ad_writer.template_to_ad_fields": {ad_field_name: template}}
        )

    def _assert_script_contains_field(self, script, name, value):
        """Assert that AD PowerShell script `script` includes the AD field
        called `name` and that the field value is `value`.
        """
        self.assertRegex(script, f'"{name}"="{value}"')

    def _get_script_contents(self, index=0):
        return self.ad_writer.scripts[index].split("\n")[5].strip()

    @parameterized.expand(
        [
            # Test without any changes to settings
            [
                dict_modifier({}),
            ],
            # Test with overridden password
            [
                dict_modifier({"primary": {"password": "Password1"}}),
            ],
            [
                dict_modifier({"primary": {"password": "Hunter2"}}),
            ],
            # Test with overridden user
            [
                dict_modifier({"primary": {"system_user": "R2D2"}}),
            ],
            [
                dict_modifier({"primary": {"system_user": "C-3PO"}}),
            ],
        ]
    )
    def test_common_ps_code(self, settings_transformer):
        """Test ps_script common code (first five lines of each script).

        The common code lines are identical for all writes to the AD.
        This is verified by the 'verify_identical_common_code' method.

        Create_user is used as a test-case to provoke the creation of a PS script,
        but only the common top of the code is actually tested.
        """
        self._setup_adwriter(settings_transformer)

        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Expected outputs
        num_expected_scripts = 3

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.create_user(mo_uuid=uuid, create_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        common_ps = self._verify_identical_common_code(num_expected_scripts)
        common_ps = [x.strip() for x in common_ps]
        expected_ps = [
            self.ad_writer._ps_boiler_plate()["encoding"],
            '$User = "' + self.settings["primary"]["system_user"] + '"',
            '$PWord = ConvertTo-SecureString –String "'
            + self.settings["primary"]["password"]
            + '" –AsPlainText -Force',
            '$TypeName = "System.Management.Automation.PSCredential"',
            (
                "$UserCredential = New-Object –TypeName $TypeName "
                "–ArgumentList $User, $PWord"
            ),
        ]
        self.assertEqual(common_ps, expected_ps)

    # Jinja to determine if mo_values['employment_number'] is a prime number
    is_prime_jinja = """
        {#- Aliased to n for readability -#}
        {%- set n = mo_values['employment_number'] | int -%}
        {#- Negative numbers, zero and one are not primes -#}
        {%- if n <= 1 -%}
            false
        {%- else -%}
            {#- Using vars as dict allows settings by reference -#}
            {%- set vars = {'is_prime': True} -%}

            {#- Really we only need to check until sqrt(n) -#}
            {%- for i in range(2,n) -%}
                {#- We cannot use percentage for modulus -#}
                {%- set quotient = (n/i) | int -%}
                {%- set remainder = n - i*quotient -%}

                {#- If it divides without remainder, it is not prime -#}
                {%- if remainder == 0 -%}
                    {#- Update is_prime by reference -#}
                    {%- set _ = vars.update({'is_prime': False}) -%}
                {%- endif -%}
            {%- endfor -%}

            {#- Print whether we found a prime or not -#}
            {%- if vars['is_prime'] -%}
                true
            {%- else -%}
                false
            {%- endif -%}
        {%- endif -%}
    """

    @parameterized.expand(
        [
            # Test without any changes
            [dict_modifier({}), None, ""],
            # Test with new employment number
            [dict_modifier({}), mo_modifier({"employment_number": "42"}), ""],
            [dict_modifier({}), mo_modifier({"employment_number": "100"}), ""],
            # Test mo_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.mo_to_ad_fields": {
                            "unit_user_key": "unit_user_key"
                        }
                    }
                ),
                None,
                '"unit_user_key"="Musik";',
            ],
            # Test with added template_field
            # Simple field lookup
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "unit_user_key": "{{ mo_values['unit_user_key'] }}"
                        }
                    }
                ),
                None,
                '"unit_user_key"="Musik";',
            ],
            # Test with timestamp as added template_field
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "extensionAttribute21": "{{ sync_timestamp }}",
                        }
                    }
                ),
                None,
                f'"extensionAttribute21"="{_SYNC_TIMESTAMP}";',
            ],
            # Field lookup and processing
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "street_number": """
                                {%- set street = mo_values['unit_streetname'] -%}
                                {{ street.split(' ')[-1] }}
                                {#- Comment used to trim whitespace -#}
                            """
                        }
                    }
                ),
                None,
                '"street_number"="451";',
            ],
            # Test if employment number is prime
            # 101 is prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "101"}),
                '"employment_number_is_prime"="true";',
            ],
            # 100 is not prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "100"}),
                '"employment_number_is_prime"="false";',
            ],
            # 263 is prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "263"}),
                '"employment_number_is_prime"="true";',
            ],
            # 267 is not prime
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "employment_number_is_prime": is_prime_jinja
                        }
                    }
                ),
                mo_modifier({"employment_number": "267"}),
                '"employment_number_is_prime"="false";',
            ],
        ]
    )
    @freeze_time(_SYNC_TIMESTAMP)
    def test_user_create_custom_fields(
        self, settings_transformer, mo_transformer, expected
    ):
        """Test user create ps_script code.

        The common code is not tested.
        """
        self._setup_adwriter(None, mo_transformer, settings_transformer)
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Expected outputs
        num_expected_scripts = 3

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.create_user(mo_uuid=uuid, create_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        create_user_ps = self._get_script_contents(index=2)

        mo_values = self.ad_writer.read_ad_information_from_mo(uuid)

        if "extensionAttribute21" in expected:
            # When checking datetime we need to drop the seconds before comparing
            expected = expected[:-11]

        expected_content = [
            "New-ADUser",
            '-Name "'
            + mo_values["full_name"]
            + " - "
            + mo_values["sam_account_name"]
            + '"',
            '-Displayname "' + mo_values["full_name"] + '"',
            '-GivenName "' + mo_values["name"][0] + '"',
            '-SurName "' + mo_values["name"][1] + '"',
            '-SamAccountName "' + mo_values["sam_account_name"] + '"',
            '-EmployeeNumber "' + mo_values["employment_number"] + '"',
            "-Credential $usercredential",
            '-UserPrincipalName "' + mo_values["sam_account_name"] + '@epn_end"',
            "-OtherAttributes",
            '"level2orgunit_field"="Ingen";',
            '"org_field"="Kommune\\Forvalting\\Enhed\\";',
            '"uuid_field"="' + mo_values["uuid"] + '";',
            '"cpr_field"="'
            + mo_values["cpr"][0:6]
            + "ad_cpr_sep"
            + mo_values["cpr"][6:]
            + '"',
            '-Path "search_base"',
            expected,
        ]
        for content in expected_content:
            self.assertIn(content, create_user_ps)

    def test_user_edit(self):
        """Test user edit ps_script code.

        The common code is not tested.
        """
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.sync_user(mo_uuid=uuid, sync_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        edit_user_ps = self._get_script_contents()
        mo_values = self.ad_writer.read_ad_information_from_mo(uuid)
        expected_content = [
            "Get-ADUser",
            "-Filter 'SamAccountName -eq \"" + mo_values["sam_account_name"] + "\"'",
            "-Credential $usercredential",
            "|",
            "Set-ADUser",
            "-Credential $usercredential",
            '-Displayname "' + mo_values["full_name"] + '"',
            '-GivenName "' + mo_values["name"][0] + '"',
            '-SurName "' + mo_values["name"][1] + '"',
            '-EmployeeNumber "' + mo_values["employment_number"] + '"',
            "-Replace",
            '"level2orgunit_field"="Ingen";',
            '"org_field"="Kommune\\Forvalting\\Enhed\\";',
        ]
        for content in expected_content:
            self.assertIn(content, edit_user_ps)

    def test_user_create_ad_values(self):
        """Test that `ad_values` is present (and empty) when creating a new AD
        user. The template `JOB_TITLE_TEMPLATE` uses the MO job title to fill
        in the job title in the appropriate AD field.
        """
        self._setup_adwriter(
            early_transform_settings=self._add_to_template_to_ad_fields(
                JOB_TITLE_AD_FIELD_NAME,
                JOB_TITLE_TEMPLATE,
            )
        )
        mo_uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.create_user(mo_uuid, create_manager=False)
        # Find the "create AD user" script
        create_script = first_true(
            self.ad_writer.scripts, pred=lambda s: "New-ADUser" in s
        )
        # Assert that it contains the mapped job title field
        self._assert_script_contains_field(
            create_script, JOB_TITLE_AD_FIELD_NAME, self.mo_values_func()["title"]
        )

    @parameterized.expand(
        [
            # Tuples of (AD job title, which value to use)
            (None, "mo-value"),
            ("", "mo-value"),
            ("Tester", "ad-value"),
        ]
    )
    def test_user_update_ad_values(self, ad_value, expectation):
        """Test that `ad_values` is present (and empty) when updating an
        existing AD user. The template `JOB_TITLE_TEMPLATE` only populates the
        AD job title with the MO job title if the current AD job title is
        empty.
        """
        self._setup_adwriter(
            early_transform_settings=self._add_to_template_to_ad_fields(
                JOB_TITLE_AD_FIELD_NAME,
                JOB_TITLE_TEMPLATE,
            ),
            transform_ad_values=dict_modifier(
                {JOB_TITLE_AD_FIELD_NAME: ad_value},
            ),
        )
        mo_uuid = "invalid-provided-and-accepted-due-to-mocking"
        mo_value = self.mo_values_func()["title"]
        self.ad_writer.sync_user(mo_uuid, ad_dump=None, sync_manager=False)
        # Find the "update AD user" script
        update_script = first_true(
            self.ad_writer.scripts, pred=lambda s: "Set-ADUser" in s
        )
        # Assert that the update script uses the current AD job title
        self._assert_script_contains_field(
            update_script,
            JOB_TITLE_AD_FIELD_NAME,
            ad_value if ad_value else mo_value,
        )

    def test_duplicate_ad_field_entries(self):
        """Test user edit ps_script code.

        The common code is not tested.
        """
        # These keys conflict, and thus, no primary_write settings are emitted
        settings_transformer = dict_modifier(
            {
                "integrations.ad_writer.template_to_ad_fields": {
                    "Name": "{{ mo_values['unit'] }}"
                },
                "integrations.ad_writer.mo_to_ad_fields": {"unit": "name"},
            }
        )
        with self.assertRaises(ValueError):
            self._setup_adwriter(early_transform_settings=settings_transformer)

    def test_non_overwritten_default(self):
        """Test that a mistake in overriding a default results in an error."""
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        # Add another mapping for the display name AD field
        # In `TestADMixin._prepare_settings.default_settings`, this is spelled
        # `Displayname` (notice the case difference.)
        display_name = {"displayname": "{{ mo_values['name'][0] }}"}
        self.settings["primary_write"]["template_to_ad_fields"].update(display_name)

        with self.assertRaises(ValueError):
            self.ad_writer.sync_user(
                mo_uuid="invalid-provided-and-accepted-due-to-mocking",
                sync_manager=False,
            )

    def test_user_edit_illegal_parameter(self):
        """Test user edit ps_script code with illegal parameter

        The common code is not tested.

        This test simply ensures that the illegal parameter is dropped.
        """
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)
        illegal_parameters["Set-ADUser"].append("Displayname")

        settings_transformer = dict_modifier({})
        self._setup_adwriter(settings_transformer)
        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.sync_user(mo_uuid=uuid, sync_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        edit_user_ps = self._get_script_contents()
        self.assertNotIn("Displayname", edit_user_ps)

    def test_user_edit_illegal_attribute(self):
        """Test user edit ps_script code with illegal attribute.

        The common code is not tested.

        This test simply ensures that the illegal attribute is dropped.
        """
        # Assert no scripts were produced from initializing ad_writer itself
        self.assertGreaterEqual(len(self.ad_writer.scripts), 0)

        settings_transformer = dict_modifier(
            {"integrations.ad_writer.template_to_ad_fields": {"Name": "John"}}
        )

        self._setup_adwriter(settings_transformer)
        # Expected outputs
        num_expected_scripts = 1

        # Run create user and fetch scripts
        uuid = "invalid-provided-and-accepted-due-to-mocking"
        self.ad_writer.sync_user(mo_uuid=uuid, sync_manager=False)
        # Check that scripts were produced
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)

        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)

        # Check that the create user ps looks good
        edit_user_ps = self._get_script_contents()
        self.assertNotIn('"Name"="John"', edit_user_ps)

    def test_duplicated_ad_field_name(self):
        # Test configuration where 'Name' is mapped both via
        # `mo_to_ad_fields` *and* `template_to_ad_fields`
        add_dupe_field = dict_modifier(
            {
                "integrations.ad_writer.mo_to_ad_fields": {"unit": "Name"},
            }
        )
        with self.assertRaises(ValueError):
            self._setup_adwriter(early_transform_settings=add_dupe_field)

    @parameterized.expand(
        [
            # Verify different employment numbers
            [
                dict_modifier({}),
                mo_modifier({"employment_number": "267"}),
                dict_modifier({}),
                [],
            ],
            [
                dict_modifier({}),
                mo_modifier({"employment_number": "42"}),
                dict_modifier({}),
                [],
            ],
            # Test mo_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.mo_to_ad_fields": {"unit": "Name"},
                    }
                ),
                mo_modifier({"employment_number": "42"}),
                dict_modifier(
                    {
                        "name": ("John Deere", "Enhed"),
                    }
                ),
                ["Name"],
            ],
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.mo_to_ad_fields": {
                            "unit": "Name",
                            "employment_number": "extension_field2",
                        },
                    }
                ),
                mo_modifier({"employment_number": "42"}),
                dict_modifier(
                    {
                        "name": ("John Deere", "Enhed"),
                        "extension_field2": (None, "42"),
                    }
                ),
                ["Name"],
            ],
            # Test template_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "adjusted_number": (
                                "{{ mo_values['employment_number']|int + 5 }}"
                            )
                        },
                    }
                ),
                mo_modifier({"employment_number": "42"}),
                dict_modifier(
                    {
                        "adjusted_number": (None, "47"),
                    }
                ),
                [],
            ],
            # Test template_to_ad_fields
            [
                dict_modifier(
                    {
                        "integrations.ad_writer.template_to_ad_fields": {
                            "adjusted_number": (
                                "{{ mo_values['employment_number']|int + 5 }}"
                            ),
                            "Enabled": "Invalid",
                        },
                    }
                ),
                mo_modifier(
                    {
                        "employment_number": "42",
                    }
                ),
                dict_modifier(
                    {
                        "adjusted_number": (None, "47"),
                        "enabled": (True, "Invalid"),
                    }
                ),
                [],
            ],
        ]
    )
    def test_sync_compare(
        self,
        settings_transformer,
        mo_transformer,
        expected_transformer,
        removed_fields,
    ):
        # If test case defines `removed_fields`, remove these fields from the
        # `integrations.ad_writer.template_to_ad_fields` setting, as they are
        # defined by the `integrations.ad_writer.mo_to_ad_fields` setting
        # instead.
        def remove_template_fields(settings):
            key = "integrations.ad_writer.template_to_ad_fields"
            to_remove = set([f.lower() for f in removed_fields])
            for field in copy.copy(settings[key]):
                if field.lower() in to_remove:
                    del settings[key][field]
            return settings

        def actual_settings_transformer(settings):
            return settings_transformer(remove_template_fields(settings))

        self._setup_adwriter(
            early_transform_settings=actual_settings_transformer,
            transform_mo_values=mo_transformer,
        )

        uuid = "invalid-provided-and-accepted-due-to-mocking"
        mo_values = self.ad_writer.read_ad_information_from_mo(uuid)
        mo_values["manager_cpr"] = None
        ad_values = self._prepare_get_from_ad(lambda x: x)
        ad_values["Name"] = "John Deere"

        def find_ad_user(cpr, ad_dump=None):
            return ad_values

        self.ad_writer._find_ad_user = find_ad_user

        mismatch = self.ad_writer._sync_compare(mo_values, None)
        expected = {
            "level2orgunit_field": (None, "Ingen"),
            "org_field": (None, "Kommune\\Forvalting\\Enhed\\"),
            "name": (ad_values["Name"], mo_values["name_sam"]),
            "displayname": (None, " ".join(mo_values["name"])),
            "givenname": (ad_values["GivenName"], mo_values["name"][0]),
            "surname": (None, mo_values["name"][1]),
            "employeenumber": (None, mo_values["employment_number"]),
        }
        expected = expected_transformer(expected)
        self.assertEqual(mismatch, expected)

        # Apply changes to AD, and check _sync_compare returns nothing
        ad_keys = ad_values.keys()
        for lower_ad_key, changeset in expected.items():
            cased_ad_key = next(
                filter(lambda ad_key: ad_key.lower() == lower_ad_key, ad_keys),
                lower_ad_key,
            )
            ad_values[cased_ad_key] = changeset[1]
        mismatch = self.ad_writer._sync_compare(mo_values, None)
        self.assertEqual(mismatch, {})

    def test_sync_compare_context_includes_ad_values(self):
        self._setup_adwriter()

        # The `ad_values` passed in the template context has all keys in lower
        # case
        ad_values = {key.lower(): val for key, val in self.ad_values_func().items()}

        mo_values = self.ad_writer.read_ad_information_from_mo("mo-uuid")
        mo_values["manager_cpr"] = None

        # Call `_sync_compare` with a mocked `_render_field_template` method
        # so we can inspect the template contexts passed to it
        with mock.patch.object(self.ad_writer, "_render_field_template") as m:
            self.ad_writer._sync_compare(mo_values, None)

        # Assert that all calls to mocked template render method includes a
        # valid `ad_values` dict in the template context
        template_contexts = [
            call.args[0] for call in m.mock_calls if len(call.args) == 2
        ]
        self.assertTrue(
            all(context["ad_values"] == ad_values for context in template_contexts)
        )

    @mock.patch("time.sleep")
    def test_rename_ad_user(self, *args):
        # Configure ADWriter to write an AD "Name" attribute consisting of the
        # user's full name, plus " testing".
        settings_transformer = dict_modifier(
            {
                "integrations.ad_writer.template_to_ad_fields": {
                    "Name": "{{ mo_values['full_name'] }} - testing"
                },
            }
        )

        # TODO: this duplicates code found in "test_mo_to_ad_sync.py".
        def remove_manager_cpr(mo_values, *args, **kwargs):
            del mo_values["manager_cpr"]
            return mo_values

        self._setup_adwriter(
            early_transform_settings=settings_transformer,
            transform_mo_values=remove_manager_cpr,
        )

        # Construct a mock AD dump, where the user has a Name attr consisting
        # of their full name only.
        ad_dump = [self._prepare_get_from_ad(ad_transformer=None)]

        # Invoke the AD rename
        mo_user_uuid = "not-really-a-uuid"
        self.ad_writer.sync_user(mo_user_uuid, ad_dump=ad_dump)

        # Assert that we issued two AD PowerShell commands: one for the rename,
        # and one for updating the other AD user attributes.
        self.assertEqual(len(self.ad_writer.scripts), 2)

        # Fetch the "rename" command itself: the last line of the first script
        rename_cmd = self.ad_writer.scripts[0].splitlines()[-1]
        # Assert that the "rename" command does indeed attempt to rename the AD
        # user to "Firstname Lastname - testing", due to the mapping in
        # `settings_transformer`.
        expected_new_name = "%s - testing" % ad_dump[0]["Name"]
        expected_cmd_fragment = '-NewName "%s"' % expected_new_name
        self.assertIn(expected_cmd_fragment, rename_cmd)

    def test_add_manager(self):
        mo_values = self.ad_writer.read_ad_information_from_mo(
            uuid="0", read_manager=True
        )
        self.ad_writer.add_manager_to_user(
            "MGORE", manager_sam=mo_values["manager_sam"]
        )
        # Expected outputs
        num_expected_scripts = 1
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)
        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)
        # Check that the create user ps looks good
        add_manager_ps = self._get_script_contents()

        expected_line = (
            "Get-ADUser -Filter 'SamAccountName -eq \"MGORE\"'"
            + " -Credential $usercredential |Set-ADUser -Manager "
            + mo_values["manager_sam"]
            + " -Credential $usercredential"
        )
        self.assertEqual(add_manager_ps, expected_line)

    def test_set_password(self):
        password = "password"
        self.ad_writer.set_user_password("MGORE", password)
        # Expected outputs
        num_expected_scripts = 1
        self.assertEqual(len(self.ad_writer.scripts), num_expected_scripts)
        # Verify that the first 4 lines are identical for all scripts
        self._verify_identical_common_code(num_expected_scripts)
        # Check that the create user ps looks good
        set_password_ps = self._get_script_contents()

        expected_line = (
            "Get-ADUser -Filter 'SamAccountName -eq \"MGORE\"' -Credential"
            + " $usercredential |Set-ADAccountPassword -Reset -NewPassword"
            + ' (ConvertTo-SecureString -AsPlainText "{}" -Force)'
            + " -Credential $usercredential"
        ).format(password)
        self.assertEqual(set_password_ps, expected_line)

    @parameterized.expand(
        [
            # AD dump is None
            (None, CprNotFoundInADException),
            # AD dump is empty list
            ([], CprNotFoundInADException),
            # AD dump has user without CPR
            ([{"foo": "bar"}], CprNotFoundInADException),
            # AD dump has exactly one matching user
            ([{"cpr": "112233-4455"}], None),
            # AD dump has multiple users with identical CPRs
            (
                [{"cpr": "112233-4455"}, {"cpr": "112233-4455"}],
                CprNotNotUnique,
            ),
        ]
    )
    def test_find_ad_user_ad_dump(self, ad_dump, expected_exception):
        cpr = "112233-4455"

        def settings_transformer(settings):
            settings["integrations.ad"][0].update({"cpr_field": "cpr"})
            return settings

        self._setup_adwriter(
            early_transform_settings=settings_transformer, mock_find_ad_user=False
        )

        if expected_exception:
            with self.assertRaises(expected_exception):
                self.ad_writer._find_ad_user(cpr, ad_dump=ad_dump)
        else:
            ad_user = self.ad_writer._find_ad_user(cpr, ad_dump=ad_dump)
            self.assertDictEqual(ad_user, {"cpr": "112233-4455"})

    def test_template_fails_on_undefined_variable(self):
        settings_transformer = dict_modifier(
            {
                "integrations.ad_writer.template_to_ad_fields": {
                    "Displayname": "{{ unknown_variable }}"
                },
            }
        )
        self._setup_adwriter(early_transform_settings=settings_transformer)
        with self.assertRaises(UndefinedError):
            self.ad_writer.sync_user(mo_uuid="mo-uuid", sync_manager=False)

    def test_fullnames_are_empty_when_constituants_are_empty(self):
        self._setup_adwriter()
        uuid = "some_uuid_here"

        def get_mo_values(firstname, surname, nickname_firstname, nickname_surname):
            self.user = {
                "uuid": "some_uuid_here",
                "navn": "some_name some_lastname",
                "efternavn": surname,
                "fornavn": firstname,
                "kaldenavn": "",
                "kaldenavn_fornavn": nickname_firstname,
                "kaldenavn_efternavn": nickname_surname,
                "cpr": "some_cpr",
            }
            self.lc = AttrDict(
                {
                    "users": {
                        self.user["uuid"]: [self.user],
                    },
                    "engagements": {
                        "engagement_uuid": [
                            {
                                "user": self.user["uuid"],
                                "primary_boolean": True,
                                "user_key": "some_userkey",
                                "job_function": "job_function_title_uuid",
                                "unit": "some_unit",
                                "uuid": "engagement_uuid",
                                "from_date": None,
                                "to_date": None,
                            }
                        ]
                    },
                    "classes": {"job_function_title_uuid": {"title": "some_job_title"}},
                    "units": {
                        "some_unit": [
                            {
                                "uuid": "some_unit",
                                "name": "some_unit_name",
                                "user_key": "some_unit_userkey",
                                "location": "some_unit_location",
                                "unit_type": "some_unit_type",
                                "level": "some_unit_level",
                                "parent": None,
                            }
                        ]
                    },
                    "addresses": {},
                    "it_connections": {
                        "it_uuid": [
                            {
                                "uuid": "it_connection_uuid",
                                "user": "user_uuid",
                                "unit": None,
                                "username": "opus_username",
                                "itsystem": "it_uuid",
                                "from_date": "1930-01-01",
                                "to_date": None,
                            }
                        ]
                    },
                }
            )
            self.lc_historic = self.lc
            self.ad_writer.lc = self.lc
            self.ad_writer.lc_historic = self.lc_historic
            self.ad_writer.datasource = LoraCacheSource(
                self.lc,
                self.lc_historic,
                MockMORESTSource(from_date=None, to_date=None),
            )
            mo_values = self.ad_writer._read_ad_information_from_mo(uuid)
            return mo_values

        mo_values = get_mo_values("Ursula", "Uniknavn", "Anne", "Jensen")
        assert isinstance(mo_values, LazyDict)
        assert mo_values["name"] == ("Ursula", "Uniknavn")
        assert mo_values["full_name"] == "Ursula Uniknavn"

        assert mo_values["nickname"] == ("Anne", "Jensen")
        assert mo_values["full_nickname"] == "Anne Jensen"

        mo_values = get_mo_values("", "", "", "")
        assert isinstance(mo_values, LazyDict)
        assert mo_values["name"] == ("", "")
        assert mo_values["full_name"] == ""

        assert mo_values["nickname"] == ("", "")
        assert mo_values["full_nickname"] == ""

        # Test all fields can be read
        for key in mo_values.keys():
            mo_values[key]

    def test_create_user_can_create_manager(self):
        # Test what happens when passing `create_manager=True` to `create_user`
        self._setup_adwriter()
        # Remember SamAccountName of user to be created
        expected_sam_account_name = self.mo_values_func()["sam_account_name"]
        # Replace `_wait_for_replication` with do-nothing stub.
        self.ad_writer._wait_for_replication = lambda sam_account_name: None
        # Create AD user and then set manager
        status, actual_sam_account_name = self.ad_writer.create_user(
            mo_uuid="mo-user-uuid", create_manager=True
        )
        # Assert we ran a PowerShell command to set user's manager
        self.assertIn("Set-ADUser -Manager", self.ad_writer.scripts[-1])
        # Assert we received the proper result from `create_user`
        self.assertTrue(status)
        self.assertEqual(actual_sam_account_name, expected_sam_account_name)

    def test_create_user_non_empty_response_is_error(self):
        # Test what happens when `create_user` encounters a non-empty response
        # from `_run_ps_script`.
        self._setup_adwriter()
        # Replace `get_ad_user` with function which always returns empty
        # response (otherwise `create_user` will raise an exception before we
        # reach the code we want to test.)
        self.ad_writer.get_from_ad = lambda **kwargs: {}
        # Replace `_run_ps_script` with function returning non-empty response
        self.ad_writer._run_ps_script = lambda ps_script: {"not": "empty"}
        # Try to create AD user
        status, msg = self.ad_writer.create_user(
            mo_uuid="mo-user-uuid", create_manager=False
        )
        self.assertFalse(status)
        self.assertRegex(msg, "Create user failed,.*")

    def test_create_user_bails_early_if_no_engagements(self):
        # Test what happens when `create_user` encounters an empty MO user
        self._setup_adwriter()
        # Replace `read_ad_information_from_mo` with function returning None
        self.ad_writer.read_ad_information_from_mo = (
            lambda mo_uuid, create_manager: None
        )
        # Assert that `create_user` bails early with an exception
        with self.assertRaises(NoPrimaryEngagementException):
            self.ad_writer.create_user(mo_uuid="mo-user-uuid", create_manager=False)

    def test_create_user_bails_early_if_ad_user_exists(self):
        # Test what happens if `create_user` encounters an already existing AD
        # user (either by username or CPR number lookup.)

        def assert_adwriter_get_ad_user_raises(matching_kwarg, expected_exception):
            self._setup_adwriter()
            # Replace `get_from_ad` with function returning True if the lookup
            # kwarg matches `matching_kwarg`.
            self.ad_writer.get_from_ad = lambda **kwargs: matching_kwarg in kwargs
            # Assert we bail early with the proper exception
            with self.assertRaises(expected_exception):
                self.ad_writer.create_user(mo_uuid="mo-user-uuid", create_manager=False)

        assert_adwriter_get_ad_user_raises("user", SamAccountNameNotUnique)
        assert_adwriter_get_ad_user_raises("cpr", CprNotNotUnique)


class TestInitNameCreator(TestCase):
    """Test `ADWriter._init_name_creator`.

    In this test, we instantiate the 'real' `ADWriter` rather than the
    `ADWriterTestSubclass` used in `TestADWriter`. This is because we want to
    test the behavior of the real `ADWriter._init_name_creator` method, rather
    than the overridden method in `ADWriterTestSubclass`.
    """

    def test_init_name_creator_reads_ad_usernames(self):
        """Calling `ADWriter` with the default args should automatically load
        occupied usernames from AD.
        """
        ad_writer = self._prepare_adwriter()
        # Assert that the single AD user mocked by `MockADParameterReader` is
        # loaded as an occupied name.
        self.assertGreater(len(ad_writer.name_creator.occupied_names), 0)
        # Assert that we loaded the usernames from AD.
        self.assertEqual(len(ad_writer.name_creator._loaded_occupied_name_sets), 1)
        self.assertIsInstance(
            ad_writer.name_creator._loaded_occupied_name_sets[0], UserNameSetInAD
        )

    def test_init_name_creator_skips_ad_usernames(self):
        """Calling `ADWriter` with `skip_occupied_names=True` should skip
        loading occupied usernames from AD as well as other sources of occupied
        usernames.
        """
        ad_writer = self._prepare_adwriter(skip_occupied_names=True)
        # Assert that no occupied usernames were loaded, and no username sets
        # were instantiated.
        self.assertSetEqual(ad_writer.name_creator.occupied_names, set())
        self.assertEqual(len(ad_writer.name_creator._loaded_occupied_name_sets), 0)

    def _prepare_adwriter(self, **kwargs):
        # Mock enough of `ADWriter` dependencies to allow it to instantiate in
        # a test.

        all_settings = {"primary": {"method": "ntlm"}, "global": mock.MagicMock()}
        path_prefix = "integrations.ad_integration"
        with mock.patch(
            f"{path_prefix}.ad_common.read_settings",
            return_value=all_settings,
        ):
            with mock.patch(
                f"{path_prefix}.ad_writer.ADWriter._create_session",
                return_value=mock.MagicMock(),
            ):
                with mock.patch(f"{path_prefix}.ad_writer.MORESTSource"):
                    with mock.patch(
                        f"{path_prefix}.user_names.ADParameterReader",
                        return_value=MockADParameterReader(),
                    ):
                        return ADWriter(**kwargs)



import unittest
from copy import deepcopy

import pytest
from more_itertools import one
from parameterized import parameterized

from integrations.SD_Lon.engagement import _is_external
from integrations.SD_Lon.engagement import (
    is_employment_id_and_no_salary_minimum_consistent,
)
from integrations.SD_Lon.tests.fixtures import get_read_employment_changed_fixture


class TestIsExternal(unittest.TestCase):
    def test_is_external_is_false_for_number_employment_id(self):
        assert not _is_external("12345")
        assert not _is_external("1")

    def test_is_external_is_true_for_non_number_employment_id(self):
        assert _is_external("ABCDE")


class TestIsEmploymentIdAndNoSalaryMinimumConsistent(unittest.TestCase):
    @parameterized.expand(
        [
            (None, "12345", 1, True),
            (9000, "External", 10000, True),
            (9000, "External", 8000, False),
            (9000, "12345", 8000, True),
            (9000, "12345", 10000, False),
        ]
    )
    def test_return_values(
        self, no_salary_minimum, employment_id, job_pos_id, expected
    ):
        engagement = one(
            get_read_employment_changed_fixture(
                employment_id=employment_id,
                job_pos_id=job_pos_id,
            )
        )["Employment"]

        assert (
            is_employment_id_and_no_salary_minimum_consistent(
                engagement, no_salary_minimum
            )
            == expected
        )

    def test_verify_job_pos_ids_identical(self):
        # Arrange
        engagement = one(get_read_employment_changed_fixture())["Employment"]

        professions = 2 * [engagement.get("Profession")]
        engagement.update({"Profession": professions})

        # Assert
        assert is_employment_id_and_no_salary_minimum_consistent(engagement, 9000)

    def test_assertion_error_when_job_pos_ids_not_identical(self):
        # Arrange
        engagement = one(get_read_employment_changed_fixture())["Employment"]

        profession1 = deepcopy(engagement.get("Profession"))

        profession2 = deepcopy(engagement.get("Profession"))
        profession2["JobPositionIdentifier"] = "1001"
        engagement.update({"Profession": [profession1, profession2]})

        # Assert
        with pytest.raises(AssertionError):
            is_employment_id_and_no_salary_minimum_consistent(engagement, 9000)



class RunDBInitException(Exception):
    pass


class RedundantForceException(Exception):
    pass


class NoNewerDumpAvailable(Exception):
    pass


class UnknownOpusAction(Exception):
    pass

class UnknownOpusUnit(Exception):
    pass

class EmploymentIdentifierNotUnique(Exception):
    pass


class ImporterrunNotCompleted(Exception):
    pass



import json
import logging

from integrations.os2sync.config import loggername

logger = logging.getLogger(loggername)


class Session:

    def raise_for_status(self):
        pass

    def get(self, *args, **kwargs):
        logger.info("GET %r %r", args, kwargs)
        return self

    def delete(self, *args, **kwargs):
        logger.info("DELETE %r %r", args, kwargs)
        return self

    def post(self, *args, **kwargs):
        logger.info("POST %r %r", args, json.dumps(kwargs))
        return self



import os
import sys
from os2mo_data_import import ImportHelper
sys.path.append('..')
import opus_import

MUNICIPALTY_NAME = os.environ.get('MUNICIPALITY_NAME', 'Opus Import')
MOX_BASE = os.environ.get('MOX_BASE', 'http://localhost:8080')
MORA_BASE = os.environ.get('MORA_BASE', 'http://localhost:80')
XML_FILE_PATH = os.environ.get('XML_FILE_PATH', '')

importer = ImportHelper(
    create_defaults=True,
    mox_base=MOX_BASE,
    mora_base=MORA_BASE,
    system_name='Opus-Import',
    end_marker='OPUS_STOP!',
    store_integration_data=True
)

# importer.new_itsystem(
#     identifier='AD',
#     system_name='Active Directory'
# )

opus = opus_import.OpusImport(
    importer,
    MUNICIPALTY_NAME,
    XML_FILE_PATH
)

opus.insert_org_units()
opus.insert_employees()
opus.add_addresses_to_employees()

importer.import_all()



import json
import requests
import click

from ra_utils.load_settings import load_settings
from pathlib import Path
from os2mo_helpers.mora_helpers import MoraHelper


# Todo: Add logger

class TestOpusConnectivity(object):
    def __init__(self):
        self.settings = load_settings()

        self.helper = MoraHelper(hostname=self.settings['mora.base'],
                                 use_cache=False)

    def _test_saml(self):
        print('Tester adgang til MO (især SAML)')
        try:
            self.helper.read_organisation()
            print(' * Kan tilgå MO uden problemer')
            print()
        except requests.exceptions.RequestException:
            print(' * Har ikke adgang til MO, check SAML eller hostnavn')
            exit(1)

    def _check_bare_minimum_keys(self):
        print('Tjeck for tilstedeværelse af konfigurationsnøgler')
        needed_keys = [
            'mora.base',
            'municipality.name',
            'integrations.opus.import.run_db',
            'integrations.opus.import.xml_path',
            'integrations.opus.eng_types_primary_order'
        ]

        missing_keys = []
        for key in needed_keys:
            if self.settings.get(key) is None:
                missing_keys.append(key)
        if missing_keys:
            print(' * Manglende nøgler: {}'.format(missing_keys))
            exit(1)
        else:
            print(' * Alle nødvendige nøgler fundet')
            print()

    def _check_xml_files_availability(self):
        print('Tjeck der findes xml-filer at importere')
        # Import will fail non-gracefully if keys are missing, hold import
        # until bare_minimum_check is performed
        from integrations.opus import opus_helpers
        dumps = opus_helpers.read_available_dumps()
        if len(dumps) == 0:
            msg = ' * Fandt ingen xml-filer i {}'
            print(msg.format(self.settings['integrations.opus.import.xml_path']))
            exit(1)
        else:
            newest_dump = sorted(dumps)[-1]
            msg = ' * Fandt {}-xml filer. Nyeste fil fra {}'
            print(msg.format(len(dumps), newest_dump))
            print()

    def _check_keys_for_diff_import(self):
        print('Tjeck for tilstedeværelse af nøgler for diff-import')
        print('Disse klasser oprettes af initial-import')
        diff_keys = [
            'mox.base',
            'opus.addresses.employee.dar',
            'opus.addresses.employee.phone',
            'opus.addresses.employee.email',
            'opus.addresses.unit.se',
            'opus.addresses.unit.cvr',
            'opus.addresses.unit.ean',
            'opus.addresses.unit.pnr',
            'opus.addresses.unit.phoneNumber',
            'opus.addresses.unit.dar',
            'integrations.opus.it_systems.ad',
            'integrations.opus.it_systems.opus'
        ]

        missing_keys = []
        for key in diff_keys:
            if self.settings.get(key) is None:
                missing_keys.append(key)
        if missing_keys:
            print(' * Manglende nøgler: {}'.format(missing_keys))
            exit(1)
        else:
            print(' * Alle nøgler for diff-import er fundet')
            print()

    def _check_lora_klasse(self, uuid, facet=None):
        url = '/klassifikation/klasse/'
        mox = self.settings['mox.base']
        response = requests.get(mox + url + uuid)
        if not response.status_code == 200:
            return False

        if facet:
            klasse = response.json()
            facet_uuid = (klasse[uuid][0]['registreringer'][0]['relationer']
                          ['facet'][0]['uuid'])
            url = '/klassifikation/facet/'
            facet_response = requests.get(mox + url + facet_uuid)
            raw_facet = facet_response.json()
            facet_titel = (
                raw_facet[facet_uuid][0]['registreringer'][0]['attributter']
                ['facetegenskaber'][0]['brugervendtnoegle']
            )
            if not facet_titel == facet:
                return False
        return True

    def _check_existense_of_base_klasser(self):
        print('Tjekker at engagements-typer for primærberegning eksisterer')
        bad_classes = []
        if not self.settings['integrations.opus.eng_types_primary_order']:
            print(' * Listen over primær-klasser er tom')
            exit(1)

        for uuid in self.settings['integrations.opus.eng_types_primary_order']:
            if not self._check_lora_klasse(uuid, facet='engagement_type'):
                bad_classes.append(uuid)

        if bad_classes:
            msg = ' * Klasse eksisterer ikke, eller ikke facet engagement_type: {}'
            print(msg.format(bad_classes))
            exit(1)
        else:
            print('* De angivne klasser til primærberegning er gyldige')
            print()

    def _check_legal_diff_klasser(self):
        print('Tjekker at de angivne adresseklasser eksisterer')
        bad_classes = []
        for uuid in [
                self.settings['opus.addresses.employee.dar'],
                self.settings['opus.addresses.employee.phone'],
                self.settings['opus.addresses.employee.email']
        ]:
            if not self._check_lora_klasse(uuid, facet='employee_address_type'):
                bad_classes.append(uuid)

        for uuid in [
                self.settings['opus.addresses.unit.se'],
                self.settings['opus.addresses.unit.cvr'],
                self.settings['opus.addresses.unit.ean'],
                self.settings['opus.addresses.unit.pnr'],
                self.settings['opus.addresses.unit.phoneNumber'],
                self.settings['opus.addresses.unit.dar']
        ]:
            if not self._check_lora_klasse(uuid, facet='org_unit_address_type'):
                bad_classes.append(uuid)

        if bad_classes:
            msg = ' * Klasse eksisterer ikke, eller har forkert facet: {}'
            print(msg.format(bad_classes))
            exit(1)
        else:
            print('* De angivne adresseklasser er gyldige')
            print()

    def _check_run_db(self, should_be_empty: bool):
        run_db = Path(self.settings['integrations.opus.import.run_db'])
        existing_run_db = run_db.is_file()

        if should_be_empty:
            print('Dette er førstegangsimport, tjek at run-db ikke findes:')
            if existing_run_db:
                print('run-db eksisterer! Skal fjernes før førstegangsimport')
                exit(1)
            else:
                print('Der findes (korrekt) ikke nogen run-db')

        else:
            print('Dette er diff-import, tjek at run-db findes:')
            if not existing_run_db:
                print('run-db eksisterer ikke! Udfør førstegangsimport!')
                exit(1)
            else:
                print(' * run-db findes som den skal')
                print()

    def _check_is_systems(self):
        print('Tjekker at opsætning af IT-systemer er korrekt')
        mo_it_systems = self.helper.read_it_systems()

        # Check opus
        found_opus = None
        opus = self.settings['integrations.opus.it_systems.opus']
        for it_system in mo_it_systems:
            if it_system['uuid'] == opus:
                found_opus = it_system['name']

        if not found_opus:
            msg = ' * Fandt ikke IT-system Opus med uuid: {}'
            print(msg.format(opus))
            exit(1)
        else:
            if found_opus.lower().find('opus') > -1:
                msg = ' * Fandt korrekt IT system til Opus: {}'
            else:
                msg = ' * Fandt IT system til Opus, men med navnet {}????'
            print(msg.format(found_opus))

        # Check ad
        found_ad = None
        ad = self.settings['integrations.opus.it_systems.ad']
        for it_system in mo_it_systems:
            if it_system['uuid'] == ad:
                found_ad = it_system['name']

        if not found_ad:
            msg = ' * Fandt ikke IT-system AD med uuid: {}'
            print(msg.format(ad))
            exit(1)
        else:
            ad_names = ['ad', 'active directory']
            name_matched = False
            for name in ad_names:
                if found_ad.lower().find(name) > -1:
                    name_matched = True
            if name_matched:
                msg = ' * Fandt korrekt IT system til AD: {}'
            else:
                msg = ' * Fandt IT system til AD, men med navnet {}????'
            print(msg.format(found_ad))
            print()

    def base_opus_check(self):
        self._test_saml()
        self._check_bare_minimum_keys()
        self._check_xml_files_availability()
        self._check_existense_of_base_klasser()

    def diff_opus_check(self):
        self._check_keys_for_diff_import()
        self._check_legal_diff_klasser()
        self._check_is_systems()
        self._check_run_db(should_be_empty=False)


@click.command(help="Test Opus configuration")
@click.option("--test-diff-import", is_flag=True)
def cli(**args):
    toc = TestOpusConnectivity()
    toc.base_opus_check()
    if args['test_diff_import']:
        toc.diff_opus_check()
    else:
        toc._check_run_db(should_be_empty=True)


if __name__ == '__main__':
    cli()



import unittest
from datetime import datetime
from pathlib import Path
from unittest.mock import MagicMock
from unittest.mock import patch
from uuid import uuid4

from hypothesis import given
from hypothesis.strategies import datetimes
from hypothesis.strategies import text
from hypothesis.strategies import uuids
from parameterized import parameterized

from integrations.opus import opus_helpers
from integrations.opus.opus_diff_import import OpusDiffImport


class OpusDiffImportTestbase(OpusDiffImport):
    @patch("integrations.calculate_primary.opus.OPUSPrimaryEngagementUpdater")
    @patch(
        "integrations.calculate_primary.opus.OPUSPrimaryEngagementUpdater.__init__",
        return_value=None,
    )
    def __init__(self, latest_date, primary_init_mock, mock_primary, *args, **kwargs):
        self.morahelper_mock = MagicMock()
        self.morahelper_mock.read_organisation.return_value = "org_uuid"
        self.morahelper_mock._mo_post.return_value.status_code = 201
        self.morahelper_mock.ensure_class_in_facet.return_code = uuid4()
        mock_primary.primary_types = {"non_primary": "test"}

        super().__init__(latest_date, *args, **kwargs)

    def _get_mora_helper(self, hostname, use_cache):
        return self.morahelper_mock

    def _find_classes(self, facet):
        if facet == "engagement_type":
            return ({"Ansat": "eng_type", "Softwaretester": "dummy"}, facet)
        if facet == "org_unit_type":
            return ({"uuid": "eng_type"}, facet)
        if facet == "org_unit_address_type":
            return (
                {
                    "SE": "SE_UUID",
                    "CVR": "CVR_UUID",
                    "EAN": "EAN_UUID",
                    "Pnummer": "PN_UUID",
                    "PhoneUnit": "Phone_UUID",
                    "AddressPostUnit": "Addr_UUID",
                },
                facet,
            )
        if facet == "employee_address_type":
            return (
                {
                    "AdressePostEmployee": "post_UUID",
                    "EmailEmployee": "email_UUID",
                    "PhoneEmployee": "phone_UUID",
                },
                facet,
            )
        return ({"uuid": "eng_type"}, facet)

    def _add_klasse_to_lora(self, klasse_name, facet_uuid):
        return self._add_klasse_to_lora


class OpusDiffImportTest_updates(OpusDiffImportTestbase):
    def __init__(self, *args, **kwargs):
        self._update_unit_types = MagicMock()
        self._update_unit_types.return_value = "DummyUUID"
        self.morahelper_mock._mo_post.return_value.status_code = 200
        self._perform_address_update = MagicMock()
        super().__init__(*args, **kwargs)

    def _update_unit_types(self, org_type):
        return self._update_unit_types

    def _perform_address_update(self, args, current, address_dict):
        return self._perform_address_update


class OpusDiffImportTest_counts(OpusDiffImportTestbase):
    def __init__(self, *args, **kwargs):
        self.update_employee = MagicMock()
        self.update_unit = MagicMock()
        self.terminate_detail = MagicMock()
        self._find_engagement = MagicMock()

        super().__init__(*args, **kwargs)

    def update_employee(self, employee):
        return self.update_employee

    def update_unit(self, unit):
        return self.update_unit

    def terminate_detail(self, uuid, detail_type="engagement", end_date=None):
        return self.terminate_detail

    def _find_engagement(self, bvn, present=False):
        return self._find_engagement


class Opus_diff_import_tester(unittest.TestCase):
    """Tests for opus_diff_import funktions that does not use input from testfiles"""

    def setUp(self):
        self.file1 = Path.cwd() / "integrations/opus/tests/ZLPE20200101_delta.xml"
        self.file2 = Path.cwd() / "integrations/opus/tests/ZLPE20200102_delta.xml"
        self.expected_unit_count = 4
        self.expected_employee_count = 3
        self.expected_terminations = 1

        filter_ids = []
        (
            units,
            filtered_units,
            employees,
            terminated_employees,
        ) = opus_helpers.read_and_transform_data(self.file1, self.file2, filter_ids)
        self.units = list(units)
        self.filtered_units = list(filtered_units)
        self.employees = list(employees)
        self.terminated_employees = list(terminated_employees)

    def test_file_diff(self):
        self.assertEqual(len(self.units), self.expected_unit_count)
        self.assertEqual(
            len(self.employees),
            self.expected_employee_count,
        )

    @given(datetimes())
    def test_import_unit_count(self, xml_date):
        self.assertIsInstance(xml_date, datetime)
        diff = OpusDiffImportTest_counts(
            xml_date, ad_reader=None, employee_mapping="test"
        )
        diff.start_import(self.units, self.employees, self.terminated_employees)
        self.assertEqual(diff.update_unit.call_count, self.expected_unit_count)

    @given(datetimes())
    def test_import_employee_count(self, xml_date):
        self.assertIsInstance(xml_date, datetime)
        diff = OpusDiffImportTest_counts(
            xml_date, ad_reader=None, employee_mapping="test"
        )
        diff.start_import(self.units, self.employees, self.terminated_employees)
        self.assertEqual(diff.update_employee.call_count, self.expected_employee_count)

    @given(datetimes())
    def test_termination(self, xml_date):
        self.assertIsInstance(xml_date, datetime)
        diff = OpusDiffImportTest_counts(
            xml_date, ad_reader=None, employee_mapping="test"
        )
        diff.start_import(self.units, self.employees, self.terminated_employees)
        self.assertEqual(
            diff._find_engagement.call_count, self.expected_terminations * 2
        )

    @patch("integrations.dawa_helper.dawa_lookup")
    @given(datetimes())
    def test_update_unit(self, dawa_helper_mock, xml_date):
        self.assertIsInstance(xml_date, datetime)
        diff = OpusDiffImportTestbase(xml_date, ad_reader=None, employee_mapping="test")
        for unit in self.units:
            diff.update_unit(unit)
            calculated_uuid = opus_helpers.generate_uuid(unit["@id"])
            if unit.get("street"):
                diff.helper._mo_post.assert_called_with(
                    "details/create",
                    {
                        "type": "address",
                        "value": dawa_helper_mock(),
                        "address_type": {
                            "uuid": str(diff.morahelper_mock.ensure_class_in_facet())
                        },
                        "validity": {"from": xml_date.strftime("%Y-%m-%d"), "to": None},
                        "org_unit": {"uuid": str(calculated_uuid)},
                        "visibility": None,
                    },
                )

    @patch("integrations.dawa_helper.dawa_lookup")
    @given(datetimes())
    def test_update_employee(self, dawa_helper_mock, xml_date):
        self.assertIsInstance(xml_date, datetime)
        diff = OpusDiffImportTestbase(xml_date, ad_reader=None, employee_mapping="test")
        diff.it_systems = {"Opus": "Opus_uuid"}
        diff.updater.primary_types = {"non_primary": "test"}
        diff.updater.set_current_person = MagicMock()
        diff.updater.recalculate_primary = MagicMock()
        with patch(
            "integrations.opus.opus_diff_import.OpusDiffImport._assert",
            return_value=None,
        ):
            for employee in self.employees:
                if employee.get("cpr"):
                    diff.update_employee(employee)
                    uuid = diff.helper._mo_lookup().__getitem__().__getitem__()
                    diff.helper._mo_post.assert_called_with(
                        "details/terminate",
                        {
                            "type": "manager",
                            "uuid": uuid,
                            "validity": {"to": xml_date.strftime("%Y-%m-%d")},
                        },
                    )
                else:
                    self.assertEqual(employee["@action"], "leave")

    @given(datetimes(), datetimes(), text(), uuids(), uuids(), uuids())
    def test_perform_address_update_create(
        self, xml_date, fromdate, value, address_type_uuid, org_unit_uuid, visibility
    ):
        diff = OpusDiffImportTestbase(
            xml_date, ad_reader=None, employee_mapping={"dummy": 1}
        )
        address_type_uuid = str(address_type_uuid)
        args = {
            "address_type": {"uuid": address_type_uuid},
            "value": value,
            "validity": {"from": fromdate.strftime("%Y-%m-%d"), "to": None},
            "unit_uuid": str(org_unit_uuid),
            "visibility": {"uuid": str(visibility)},
        }
        current = {"uuid": address_type_uuid}
        # Same
        diff._perform_address_update(args, current)
        diff.helper.assert_not_called()
        # new
        diff._perform_address_update(args, None)
        diff.helper._mo_post.assert_called_with(
            "details/create",
            {
                "type": "address",
                "value": value,
                "address_type": {"uuid": address_type_uuid},
                "validity": {"from": fromdate.strftime("%Y-%m-%d"), "to": None},
                "org_unit": {"uuid": str(org_unit_uuid)},
                "visibility": {"uuid": str(visibility)},
            },
        )

    @parameterized.expand(
        [
            (None, None, None),
            (None, "Username", "details/create"),
            ("Username", "new_username", "details/edit"),
            ("new_username", None, "details/terminate"),
        ]
    )
    @patch("integrations.dawa_helper.dawa_lookup")
    @given(datetimes())
    def test_update_username(
        self,
        current_username,
        new_username,
        change_type,
        dawa_helper_mock,
        xml_date,
    ):
        diff = OpusDiffImportTestbase(xml_date, ad_reader=None, employee_mapping="test")
        date = xml_date.strftime("%Y-%m-%d")
        diff.it_systems = {"Opus": "Opus_uuid"}
        diff.morahelper_mock.get_e_itsystems.return_value = [
            {"user_key": current_username, "uuid": "dummyuuid"}
            if current_username
            else None
        ]

        if change_type == "details/edit":
            expected = {
                "type": "it",
                "uuid": "dummyuuid",
                "data": {
                    "user_key": new_username,
                    "validity": {"from": date, "to": None},
                },
            }
        elif change_type == "details/create":
            expected = {
                "type": "it",
                "user_key": new_username,
                "itsystem": {"uuid": "Opus_uuid"},
                "person": {"uuid": "dummyuuid"},
                "validity": {"from": date, "to": None},
            }
        elif change_type == "details/terminate":
            expected = {
                "type": "it",
                "uuid": "dummyuuid",
                "validity": {"to": date},
            }
            diff.morahelper_mock._mo_post.return_value.status_code = 200

        diff.connect_it_system(new_username, "Opus", {}, "dummyuuid")
        if change_type:
            diff.morahelper_mock._mo_post.assert_called_once_with(change_type, expected)
        else:
            diff.morahelper_mock._mo_post.assert_not_called()

    @patch("integrations.dawa_helper.dawa_lookup")
    @given(datetimes())
    def test_skip_multiple_usernames(
        self,
        dawa_helper_mock,
        xml_date,
    ):
        diff = OpusDiffImportTestbase(xml_date, ad_reader=None, employee_mapping="test")
        date = xml_date.strftime("%Y-%m-%d")
        diff.it_systems = {"Opus": "Opus_uuid"}
        diff.morahelper_mock.get_e_itsystems.return_value = [
            {"user_key": "username1", "uuid": "dummyuuid"},
            {"user_key": "username2", "uuid": "dummyuuid"},
        ]
        with self.assertLogs() as cm:
            diff.connect_it_system("new_username", "Opus", {}, "personuuid")
            diff.morahelper_mock._mo_post.assert_not_called()
            assert cm.output == ['WARNING:opusImport:Skipped connecting Opus IT system . More than one '"IT system found for person_uuid='personuuid'"]


if __name__ == "__main__":
    unittest.main()



import uuid
from datetime import date
from datetime import datetime
from functools import partial
from itertools import chain
from operator import itemgetter
from typing import Iterable
from typing import List
from typing import Union

import los_files
import payloads as mo_payloads
import util
from more_itertools import partition
from pydantic import BaseModel
from pydantic import Field
from ra_utils.generate_uuid import uuid_generator


class _ManagerBase(BaseModel):
    org_uuid: uuid.UUID = Field(alias="OrgUUID")
    cpr: str = Field(alias="CPR")  # CPR for associated employee


class ManagerCreate(_ManagerBase):
    manager_type_uuid: uuid.UUID = Field(alias="LedertypeUUID")
    manager_level_uuid: uuid.UUID = Field(alias="LederniveauUUID")
    responsibility_uuid: uuid.UUID = Field(alias="LederansvarUUID")


class ManagerEdit(ManagerCreate):
    # CSV schema is identical to `ManagerCreate`
    pass


class ManagerTerminate(_ManagerBase):
    # CSV schema is identical to `_ManagerBase`
    pass


class ManagerImporter:
    def __init__(self):
        self.cpr_cache = {}
        self.uuid_generator = uuid_generator("AAK")

    def _generate_rel_uuid(self, manager):
        return self.uuid_generator(f"{manager.org_uuid}{manager.cpr}manager")

    def cache_cpr(self):
        """Read all employees from OS2mo and cache them based on their CPR no."""
        print("Caching employees")
        employees = util.lookup_employees()
        # Not really a cache - but a mapping from the CPR found in the CSV file
        # to the MO user's UUID. Needed in `generate_manager_payload`.
        cache = map(
            itemgetter("cpr_no", "uuid"),
            filter(lambda employee: employee.get("cpr_no"), employees),
        )
        self.cpr_cache = dict(cache)

    def generate_manager_payload(
        self, manager: Union[ManagerCreate, ManagerEdit]
    ) -> dict:
        person_uuid = self.cpr_cache.get(manager.cpr)
        if not person_uuid:
            print(f"No person found for CPR {manager.cpr}")

        return mo_payloads.create_manager(
            uuid=self._generate_rel_uuid(manager),
            person_uuid=person_uuid,
            org_unit_uuid=manager.org_uuid,
            responsibility_uuid=manager.responsibility_uuid,
            manager_level_uuid=manager.manager_level_uuid,
            manager_type_uuid=manager.manager_type_uuid,
            from_date=date.today().isoformat(),
            to_date=None,
        )

    def create_manager_payloads(
        self, managers: List[Union[ManagerCreate, ManagerEdit]]
    ) -> Iterable[dict]:
        return map(self.generate_manager_payload, managers)

    async def handle_create(self, filename: str, filedate: datetime):
        """
        Handle creating new manager functions
        """
        managers = los_files.read_csv(filename, ManagerCreate)
        manager_payloads = self.create_manager_payloads(managers)  # type: ignore
        async with util.get_client_session() as session:
            await util.create_details(session, manager_payloads)

    async def handle_edit(self, filename: str, filedate: datetime):
        """
        Handle changes to managers
        """
        managers = los_files.read_csv(filename, ManagerEdit)
        manager_payloads = self.create_manager_payloads(managers)  # type: ignore
        orgfunk_uuids = set(await util.lookup_organisationfunktion())
        detail_creates, detail_edits = partition(
            lambda payload: payload["uuid"] in orgfunk_uuids, manager_payloads
        )
        converter = partial(
            mo_payloads.convert_create_to_edit, from_date=filedate.date().isoformat()
        )
        edits = map(converter, chain(manager_payloads, detail_edits))
        async with util.get_client_session() as session:
            await util.create_details(session, detail_creates)
            await util.edit_details(session, edits)

    async def handle_terminate(self, filename: str, filedate: datetime):
        """
        Handle termination of managers
        """
        managers = los_files.read_csv(filename, ManagerTerminate)
        payloads = [
            mo_payloads.terminate_detail(
                "manager",
                self._generate_rel_uuid(manager),
                filedate,
            )
            for manager in managers
        ]
        async with util.get_client_session() as session:
            await util.terminate_details(session, payloads, ignored_http_statuses=None)

    async def run(self, last_import: datetime):
        print("Starting manager import")
        filenames = los_files.get_fileset_implementation().get_import_filenames()

        self.cache_cpr()

        creates = los_files.parse_filenames(
            filenames, prefix="Leder_nye", last_import=last_import
        )
        edits = los_files.parse_filenames(
            filenames, prefix="Leder_ret", last_import=last_import
        )
        terminates = los_files.parse_filenames(
            filenames, prefix="Leder_luk", last_import=last_import
        )

        for filename, filedate in creates:
            await self.handle_create(filename, filedate)

        for filename, filedate in edits:
            await self.handle_edit(filename, filedate)

        for filename, filedate in terminates:
            await self.handle_terminate(filename, filedate)

        print("Manager import done")



#!/usr/bin/env python3
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
from datetime import date
from functools import lru_cache
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

from pydantic import AnyHttpUrl
from pydantic import BaseSettings
from pydantic import conint
from pydantic import Extra
from pydantic import Field
from pydantic import PositiveInt
from pydantic import SecretStr
from pydantic import UUID4
from ra_utils.load_settings import load_settings

from integrations.SD_Lon.models import JobFunction


class CommonSettings(BaseSettings):
    """
    Settings common to both the SD importer and SD-changed-at
    """

    mora_base: AnyHttpUrl = Field("http://mo-service:5000")
    mox_base: AnyHttpUrl = Field("http://mox-service:8080")
    sd_employment_field: Optional[str] = Field(default=None, regex="extension_[0-9]+")
    sd_global_from_date: date
    sd_import_too_deep: List[str] = []
    sd_institution_identifier: str
    sd_password: SecretStr
    sd_user: str
    sd_job_function: JobFunction
    sd_monthly_hourly_divide: PositiveInt


def gen_json_file_settings_func(settings_class: Type[CommonSettings]):
    def json_file_settings(settings: BaseSettings) -> Dict[str, Any]:
        try:
            json_settings = load_settings()
        except FileNotFoundError:
            return dict()

        # Remove the "integrations.SD_Lon." part of the key name
        json_settings = {
            key.replace("integrations.SD_Lon.", "sd_"): value
            for key, value in json_settings.items()
        }

        # Remove any double "sd_sd_" in the keys
        json_settings = {
            key.replace("sd_sd_", "sd_"): value for key, value in json_settings.items()
        }

        # Replace dots with underscores to be Pydantic compliant
        json_settings = {
            key.replace(".", "_"): value for key, value in json_settings.items()
        }

        # Remove settings forbidden according to the Settings model
        properties = settings_class.schema()["properties"].keys()
        json_settings = {
            key: value for key, value in json_settings.items() if key in properties
        }

        return json_settings

    return json_file_settings


class ImporterSettings(CommonSettings):
    municipality_code: conint(ge=100, le=999)  # type: ignore
    municipality_name: str
    sd_import_run_db: str
    sd_importer_create_associations: bool = True
    sd_importer_create_email_addresses: bool = True
    sd_importer_employment_date_as_engagement_start_date: bool = False
    sd_no_salary_minimum_id: Optional[int] = None
    sd_skip_employment_types: List[str] = []
    sd_skip_job_functions: List[str] = []
    sd_terminate_engagement_with_to_only: bool = True
    sd_use_ad_integration: bool = True

    class Config:
        extra = Extra.forbid

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                gen_json_file_settings_func(ImporterSettings),
                file_secret_settings,
            )


class ChangedAtSettings(CommonSettings):
    sd_fix_departments_root: Optional[UUID4] = None
    sd_import_run_db: str
    sd_no_salary_minimum_id: Optional[int] = None
    sd_skip_job_functions: List[str] = []
    sd_terminate_engagement_with_to_only: bool = True
    sd_use_ad_integration: bool = True

    class Config:
        extra = Extra.forbid

        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                gen_json_file_settings_func(ChangedAtSettings),
                file_secret_settings,
            )


@lru_cache()
def get_importer_settings(*args, **kwargs) -> ImporterSettings:
    return ImporterSettings(*args, **kwargs)


@lru_cache()
def get_changed_at_settings(*args, **kwargs) -> ChangedAtSettings:
    return ChangedAtSettings(*args, **kwargs)



from unittest import TestCase

from integrations.SD_Lon.config import ChangedAtSettings
from integrations.SD_Lon.sync_job_id import JobIdSync


class JobIdSyncTest(JobIdSync):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _read_classes(self):
        self.engagement_types = []
        self.job_function_types = []


class Test_sync_job_id(TestCase):
    def setUp(self):
        settings = ChangedAtSettings(
            mora_base="http://dummy.url",
            sd_job_function="JobPositionIdentifier",
            sd_import_run_db="run_db.sqlite",
            sd_institution_identifier="XY",
            sd_monthly_hourly_divide=9000,
            sd_password="secret",
            sd_user="user",
            sd_global_from_date="2022-01-09",
        )
        self.job_id_sync = JobIdSyncTest(settings)

    def test_create(self):
        self.assertTrue(self.job_id_sync.update_job_functions)



import uuid
from datetime import datetime
from functools import partial
from itertools import chain
from typing import Optional

import config
import los_files
import payloads as mo_payloads
import pydantic
import util
import uuids
from more_itertools import bucket
from more_itertools import first
from more_itertools import partition
from pydantic import Field
from ra_utils.generate_uuid import uuid_generator


class Person(pydantic.BaseModel):
    person_uuid: uuid.UUID = Field(alias="UUID")
    bvn: str = Field(alias="BrugervendtNøgle")
    cpr: str = Field(alias="CPR")
    name: str = Field(alias="CPRNavn")
    start_time: datetime = Field(alias="StartTid")
    end_time: datetime = Field(alias="SlutTid")
    email: Optional[str] = Field(alias="Email")
    primary: str = Field(alias="PrimærAnsættelse")
    org_unit_uuid: uuid.UUID = Field(alias="OrgenhedUUID")
    engagement_type_uuid: uuid.UUID = Field(alias="EngagementTypeUUID")
    az_id: Optional[str] = Field(alias="AZID")
    ad_it_system_uuid: Optional[uuid.UUID] = Field(alias="ADITSystemUUID")
    ad_account: Optional[str] = Field(alias="ADkonto")
    ad_acccount_uuid: Optional[uuid.UUID] = Field(alias="ADkontoUUID")
    job_function_uuid: uuid.UUID = Field(alias="StillingsBetegnelseUUID")


class PersonTermination(pydantic.BaseModel):
    engagement_uuid: uuid.UUID = Field(alias="UUID")
    cpr: str = Field(alias="CPR")
    name: str = Field(alias="CPRNavn")


class PersonImporter:
    def __init__(self):
        self.uuid_generator = uuid_generator("AAK")
        self.settings = config.get_config()

    def generate_employee_email_payload(self, person: Person, filedate: str):
        # For creation of Employee address/IT details we always
        # use the parsed file-date as from-date
        # The start_time field is still used for creating engagements
        _, to_date = util.convert_validities(
            person.start_time.date(), person.end_time.date()
        )
        return mo_payloads.create_address(
            uuid=self.uuid_generator(person.cpr + "email"),
            value=person.email,
            from_date=filedate,
            to_date=to_date,
            person_uuid=self.uuid_generator(person.cpr),
            address_type_uuid=uuids.PERSON_EMAIL,
        )

    def generate_employee_az_id_payload(self, person: Person, filedate: str):
        # For creation of Employee address/IT details we always
        # use the parsed file-date as from-date
        # The start_time field is still used for creating engagements
        _, to_date = util.convert_validities(
            person.start_time.date(), person.end_time.date()
        )
        return mo_payloads.create_it_rel(
            uuid=self.uuid_generator(person.cpr + "azid"),
            user_key=person.az_id,
            person_uuid=self.uuid_generator(person.cpr),
            itsystem_uuid=self.settings.azid_it_system_uuid,
            from_date=filedate,
            to_date=to_date,
        )

    def generate_employee_ad_account_payload(self, person: Person, filedate: str):
        # For creation of Employee address/IT details we always
        # use the parsed file-date as from-date
        # The start_time field is still used for creating engagements
        _, to_date = util.convert_validities(
            person.start_time.date(), person.end_time.date()
        )
        return mo_payloads.create_it_rel(
            uuid=person.ad_acccount_uuid,
            user_key=person.ad_account,
            person_uuid=self.uuid_generator(person.cpr),
            itsystem_uuid=person.ad_it_system_uuid,
            from_date=filedate,
            to_date=to_date,
        )

    def generate_engagement_payload(self, person: Person):
        from_date, to_date = util.convert_validities(
            person.start_time.date(), person.end_time.date()
        )

        primary_uuids = {"Nej": uuids.NOT_PRIMARY, "Ja": uuids.PRIMARY}
        return mo_payloads.create_engagement(
            uuid=person.person_uuid,
            org_unit_uuid=person.org_unit_uuid,
            person_uuid=self.uuid_generator(person.cpr),
            job_function_uuid=person.job_function_uuid,
            engagement_type_uuid=person.engagement_type_uuid,
            primary_uuid=primary_uuids[person.primary],
            user_key=person.bvn,
            from_date=from_date,
            to_date=to_date,
        )

    def generate_employee_payload(self, person: Person):
        # Same person info in every row
        return mo_payloads.create_employee(
            name=person.name, cpr_no=person.cpr, uuid=self.uuid_generator(person.cpr)
        )

    @staticmethod
    def generate_engagement_termination_payload(
        termination: PersonTermination, to_date: datetime
    ):
        return mo_payloads.terminate_detail(
            "engagement", termination.engagement_uuid, to_date
        )

    def create_employee_payloads(self, persons):
        cpr_buckets = bucket(persons, key=lambda x: x.cpr)

        # Every person row contains the same info, so we just pick one
        unique_persons = map(lambda key: first(cpr_buckets[key]), cpr_buckets)
        return map(self.generate_employee_payload, unique_persons)

    def create_detail_payloads(self, persons, filedate: str):
        engagement_payloads = map(self.generate_engagement_payload, persons)
        email_payloads = map(
            partial(self.generate_employee_email_payload, filedate=filedate),
            filter(lambda person: person.email, persons),
        )
        az_id_payloads = map(
            partial(self.generate_employee_az_id_payload, filedate=filedate),
            filter(lambda person: person.az_id, persons),
        )
        ad_account_payloads = map(
            partial(self.generate_employee_ad_account_payload, filedate=filedate),
            filter(lambda person: person.ad_account, persons),
        )
        return chain(
            engagement_payloads,
            email_payloads,
            az_id_payloads,
            ad_account_payloads,
        )

    async def handle_create(self, filename: str, filedate: datetime):
        """
        Handle creating new persons and details
        We are guaranteed to only have one row per person
        """
        persons = los_files.read_csv(filename, Person)
        employee_payloads = self.create_employee_payloads(persons)
        detail_payloads = self.create_detail_payloads(
            persons, filedate.date().isoformat()
        )

        async with util.get_client_session() as session:
            await util.create_details(
                session, chain(employee_payloads, detail_payloads)
            )

    async def handle_edit(self, filename: str, filedate: datetime):
        """
        Handle changes to existing persons and details
        We are guaranteed to only have one row per person

        New details on an existing person will show up in this file, rather than the
        'nye' file. So we have to potentially perform inserts of new data.

        As a row contains information about the person as well as its details,
        we do not know what has been changed. However, all information is managed
        by the external system so we can safely reimport the "same" data, as opposed to
        trying to compare the existing objects in OS2mo
        """
        persons = los_files.read_csv(filename, Person)
        employee_payloads = self.create_employee_payloads(persons)
        detail_payloads = self.create_detail_payloads(
            persons, filedate.date().isoformat()
        )

        orgfunk_uuids = set(await util.lookup_organisationfunktion())
        detail_creates, detail_edits = partition(
            lambda payload: payload["uuid"] in orgfunk_uuids, detail_payloads
        )
        converter = partial(
            mo_payloads.convert_create_to_edit, from_date=filedate.date().isoformat()
        )
        edits = map(converter, chain(employee_payloads, detail_edits))

        async with util.get_client_session() as session:
            await util.create_details(session, detail_creates)
            await util.edit_details(session, edits)

    async def handle_terminate(self, filename: str, filedate: datetime):
        """
        Handle termination of engagements. We are guaranteed one row per engagement.
        """
        persons = los_files.read_csv(filename, PersonTermination)
        termination_fn = partial(
            self.generate_engagement_termination_payload, to_date=filedate
        )
        payloads = map(termination_fn, persons)

        async with util.get_client_session() as session:
            await util.terminate_details(session, payloads)

    async def run(self, last_import: datetime):
        print("Starting person import")
        filenames = los_files.get_fileset_implementation().get_import_filenames()

        creates = los_files.parse_filenames(
            filenames, prefix="Pers_nye", last_import=last_import
        )
        edits = los_files.parse_filenames(
            filenames, prefix="Pers_ret", last_import=last_import
        )
        terminates = los_files.parse_filenames(
            filenames, prefix="Pers_luk", last_import=last_import
        )

        for filename, filedate in creates:
            await self.handle_create(filename, filedate)

        for filename, filedate in edits:
            await self.handle_edit(filename, filedate)

        for filename, filedate in terminates:
            await self.handle_terminate(filename, filedate)

        print("Person import done")



import json
import logging
import os
from pathlib import Path
from typing import Any, Optional, Dict, List
from uuid import UUID

from pydantic import BaseSettings, AnyHttpUrl
from ra_utils.apply import apply
from ra_utils.headers import TokenSettings

# flake8: noqa
from ra_utils.load_settings import load_settings

from functools import lru_cache


def json_config_settings_source(settings: BaseSettings) -> Dict[str, Any]:
    """
    Read config from settings.json.

    Reads all keys starting with 'os2sync.' and a few common settings into Settings.
    """
    try: 
        all_settings = load_settings()
    except FileNotFoundError:
        #No settingsfile found. Using environment variables"
        return {}
    #Read os2sync specific settings
    os2sync_settings = dict(
        filter(
            apply(lambda key, value: key.startswith("os2sync")), all_settings.items()
        )
    )
    
    # replace dots with underscore. eg: os2sync.ignored.unit_levels -> os2sync_ignored_unit_levels
    final_settings = {key.replace(".", "_"): val for key, val in os2sync_settings.items()}

    #Add needed common settings
    municipality = all_settings.get("municipality.cvr")
    if municipality:
        final_settings["municipality"] = municipality
    mora_base = all_settings.get("mora.base")
    if mora_base:
        final_settings["mora_base"] = mora_base

    return final_settings


class Settings(BaseSettings):
    #common:
    municipality: str   #Called "municipality.cvr" in settings.json
    mora_base: AnyHttpUrl = "http://localhost:5000"     #"mora.base" from settings.json + /service
    
    #os2sync:
    os2sync_top_unit_uuid: UUID
    os2sync_api_url: AnyHttpUrl = "http://localhost:8081"

    os2sync_use_lc_db: bool = False
    
    os2sync_log_level: int = logging.INFO
    os2sync_log_file: Path = "/opt/dipex"
    os2sync_hash_cache: Path = "/opt/dipex"
    os2sync_xfer_cpr: bool = False
    
    os2sync_autowash: bool = False
    os2sync_ca_verify_os2sync: bool = True
    os2sync_ca_verify_os2mo: bool = True

    os2sync_phone_scope_classes: List[UUID] = []
    os2sync_email_scope_classes: List[UUID] = []
    os2sync_ignored_unit_levels: List[UUID] = []
    os2sync_ignored_unit_types: List[UUID] = []
    os2sync_templates: Dict = {}
    
    os2sync_sync_managers: bool = False
    os2sync_use_contact_for_tasks: bool = False
    os2sync_employee_engagement_address: List[str] = []

    
    os2sync_truncate_length: int = 200

    class Config:
        
        env_file_encoding = "utf-8"
        @classmethod
        def customise_sources(
            cls,
            init_settings,
            env_settings,
            file_secret_settings,
        ):
            return (
                init_settings,
                env_settings,
                json_config_settings_source,
                file_secret_settings,
            )

@lru_cache()
def get_os2sync_settings(*args, **kwargs) -> Settings:
    return Settings(*args, **kwargs)

logformat = "%(levelname)s %(asctime)s %(name)s %(message)s"
loggername = "os2sync"

if __name__ == "__main__":
    print(get_os2sync_settings())


#
# Copyright (c) 2018, Magenta ApS
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import logging
from typing import Optional

from more_itertools import flatten
from more_itertools import only
from sqlalchemy.orm import sessionmaker

from constants import AD_it_system
from exporters.sql_export.lc_for_jobs_db import get_engine  # noqa
from exporters.sql_export.sql_table_defs import Adresse
from exporters.sql_export.sql_table_defs import Bruger
from exporters.sql_export.sql_table_defs import Engagement
from exporters.sql_export.sql_table_defs import Enhed
from exporters.sql_export.sql_table_defs import ItForbindelse
from exporters.sql_export.sql_table_defs import ItSystem
from exporters.sql_export.sql_table_defs import KLE
from exporters.sql_export.sql_table_defs import Leder
from integrations.os2sync.config import get_os2sync_settings
from integrations.os2sync import config
from integrations.os2sync import os2mo
from integrations.os2sync.templates import Person
from integrations.os2sync.templates import User

logger = logging.getLogger(config.loggername)


def get_session(engine):
    return sessionmaker(bind=engine, autoflush=False)()


def has_kle():
    # nothing in db if no kle
    return True


scope_to_scope = {
    "E-mail": "EMAIL",
    "Telefon": "PHONE",
    "DAR": "DAR",
    "EAN": "EAN",
    "P-nummer": "PNUMBER",
    "Text": "TEXT",
    "Ansvarlig": "ANSVARLIG",
    "Udførende": "UDFOERENDE",
    "Indsigt": "INDSIGT",
}


def try_get_ad_user_key(session, uuid: str) -> Optional[str]:
    ad_system_user_names = (
        session.query(ItForbindelse.brugernavn)
        .join(ItSystem, ItForbindelse.it_system_uuid == ItSystem.uuid)
        .filter(ItSystem.navn == AD_it_system, ItForbindelse.bruger_uuid == uuid)
        .all()
    )
    ad_system_user_names = list(flatten(ad_system_user_names))

    if len(ad_system_user_names) != 1:
        return
    return ad_system_user_names[0]


def to_mo_employee(employee):
    """Convert `Bruger` row `employee` to something which resembles a MO
    employee JSON response.

    This is done so we can pass a suitable template context to
    `os2sync.templates.Person` even when running with `OS2SYNC_USE_LC_DB=True`.
    """

    def or_none(val):
        return val or None

    def to_name(*parts):
        return or_none(" ".join(part for part in parts if part))

    return dict(
        # Name
        name=to_name(employee.fornavn, employee.efternavn),
        givenname=or_none(employee.fornavn),
        surname=or_none(employee.efternavn),
        # Nickname
        nickname=to_name(employee.kaldenavn_fornavn, employee.kaldenavn_efternavn),
        nickname_givenname=or_none(employee.kaldenavn_fornavn),
        nickname_surname=or_none(employee.kaldenavn_efternavn),
        # Other fields
        cpr_no=or_none(employee.cpr),
        user_key=or_none(employee.bvn),
        uuid=or_none(employee.uuid),
    )


def get_sts_user(session, uuid, allowed_unitids, settings):
    employee = session.query(Bruger).filter(Bruger.uuid == uuid).one()
    user = User(
        dict(
            uuid=uuid,
            candidate_user_id=try_get_ad_user_key(session, uuid),
            person=Person(to_mo_employee(employee), settings=settings),
        ),
        settings=settings,
    )

    sts_user = user.to_json()

    addresses = []
    for lc_address in session.query(Adresse).filter(Adresse.bruger_uuid == uuid).all():
        address = {
            "address_type": {
                "uuid": lc_address.adressetype_uuid,
                "scope": scope_to_scope[lc_address.adressetype_scope],
            },
            "name": lc_address.værdi,
            "value": lc_address.dar_uuid,
            "uuid": lc_address.uuid,  # not used currently
        }
        addresses.append(address)
    os2mo.addresses_to_user(sts_user, addresses, phone_scope_classes=settings.os2sync_phone_scope_classes, email_scope_classes=settings.os2sync_email_scope_classes)

    engagements = []
    for lc_engagement in (
        session.query(Engagement).filter(Engagement.bruger_uuid == uuid).all()
    ):
        engagements.append(
            {
                "uuid": lc_engagement.uuid,
                "org_unit": {"uuid": lc_engagement.enhed_uuid},
                "job_function": {"name": lc_engagement.stillingsbetegnelse_titel},
                "is_primary": lc_engagement.primær_boolean,
            }
        )

    os2mo.engagements_to_user(sts_user, engagements, allowed_unitids)

    return sts_user


top_per_unit = {}


def get_top_unit(session, lc_enhed):
    """
    return the top unit for a unit
    """
    top_unit = top_per_unit.get(lc_enhed.uuid)
    if top_unit:
        return top_unit
    branch = [lc_enhed.uuid]

    # walk as far up as necessary
    while lc_enhed.forældreenhed_uuid is not None:
        uuid = lc_enhed.forældreenhed_uuid
        top_unit = top_per_unit.get(uuid)
        if top_unit:
            break
        branch.append(uuid)
        lc_enhed = session.query(Enhed).filter(Enhed.uuid == uuid).one()
        top_unit = uuid  # last one effective

    # register top unit for all encountered
    for buuid in branch:
        top_per_unit[buuid] = top_unit
    return top_unit


def is_ignored(unit, settings):
    """Determine if unit should be left out of transfer

    Example:
        >>> from unittest.mock import Mock
        >>> unit=Mock(enhedsniveau_uuid="1", enhedstype_uuid="2")
        >>> settings={
        ... "ignored_unit_levels": ["10","2"],
        ... "ignored_unit_types":['6','7']}
        >>> is_ignored(unit, settings)
        False
        >>> unit.enhedstype_uuid="6"
        >>> is_ignored(unit, settings)
        True
        >>> unit.enhedstype_uuid="2"
        >>> is_ignored(unit, settings)
        False
        >>> unit.enhedsniveau_uuid="2"
        >>> is_ignored(unit, settings)
        True

    Args:
        unit: The organization unit to enrich with kle information.
        settings: a dictionary

    Returns:
        Boolean
    """

    return (
        unit.enhedstype_uuid in settings.os2sync_ignored_unit_types or
        unit.enhedsniveau_uuid in settings.os2sync_ignored_unit_levels)


def get_sts_orgunit(session, uuid):
    base = session.query(Enhed).filter(Enhed.uuid == uuid).one()

    if is_ignored(base, settings):
        logger.info(
            "Ignoring %s (%s, %s)",
            base.uuid,
            base.enhedsniveau_titel,
            base.enhedstype_titel,
        )
        return None

    top_unit = get_top_unit(session, base)
    if top_unit != settings.os2sync_top_unit_uuid:
        logger.debug(f"ignoring unit {uuid=}, as it is not a unit bellow {top_unit_uuid=}")
        return None

    sts_org_unit = {"ItSystemUuids": [], "Name": base.navn, "Uuid": uuid}

    if base.forældreenhed_uuid is not None:
        sts_org_unit["ParentOrgUnitUuid"] = base.forældreenhed_uuid

    itconnections = (
        session.query(ItForbindelse).filter(ItForbindelse.enhed_uuid == uuid).all()
    )
    os2mo.itsystems_to_orgunit(
        sts_org_unit,
        [{"itsystem": {"uuid": itf.it_system_uuid}} for itf in itconnections],
    )

    addresses = []
    for lc_address in session.query(Adresse).filter(Adresse.enhed_uuid == uuid).all():
        address = {
            "address_type": {
                "uuid": lc_address.adressetype_uuid,
                "user_key": lc_address.adressetype_bvn,
                "scope": scope_to_scope[lc_address.adressetype_scope],
            },
            "name": lc_address.værdi,
            "value": lc_address.dar_uuid,
            "uuid": lc_address.uuid,  # not used currently
        }
        addresses.append(address)
    os2mo.addresses_to_orgunit(sts_org_unit, addresses)

    if settings.os2sync_sync_managers:
        lc_manager = session.query(Leder).filter(Leder.enhed_uuid == uuid).all()
        manager_uuid = only(lc_manager.bruger_uuid)
        sts_org_unit.update({'managerUuid': manager_uuid})

    mokles = {}
    lc_kles = session.query(KLE).filter(KLE.enhed_uuid == uuid).all()
    for lc_kle in lc_kles:
        mokles[lc_kle.uuid] = {
            "kle_number": {"uuid": lc_kle.kle_nummer_uuid},
        }
    os2mo.kle_to_orgunit(sts_org_unit, mokles.values())
    os2mo.strip_truncate_and_warn(sts_org_unit, sts_org_unit)

    return sts_org_unit



import logging
from functools import lru_cache
from functools import partial
from functools import wraps
from operator import itemgetter
from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Tuple

import click
from ra_utils.apply import apply
from ra_utils.catchtime import catchtime
from ra_utils.jinja_filter import create_filters
from ra_utils.lazy_dict import LazyDict
from ra_utils.lazy_dict import LazyEval
from ra_utils.lazy_dict import LazyEvalBare
from ra_utils.load_settings import load_settings
from tqdm import tqdm

from .ad_exceptions import NoActiveEngagementsException
from .ad_exceptions import NoPrimaryEngagementException
from .ad_logger import start_logging
from .ad_reader import ADParameterReader
from .ad_writer import ADWriter
from exporters.sql_export.lora_cache import LoraCache

logger = logging.getLogger("CreateAdUsers")

FilterFunction = Callable[[Tuple[Dict, Dict]], bool]


class AdLifeCycle:
    def __init__(
        self, read_from_cache: bool = True, skip_occupied_names_check: bool = False
    ) -> None:
        logger.info("AD Sync Started")
        self._settings = self._load_settings()

        self.roots = self._settings["integrations.ad.write.create_user_trees"]

        self.stats = self._gen_stats()

        self.create_filters = self._load_jinja_filters("create_filters")
        self.disable_filters = self._load_jinja_filters("disable_filters")

        self.ad_reader = self._get_adreader()

        # This is a potentially slow step (since it may read LoraCache)
        print("Retrive LoRa dump")
        with catchtime() as t:
            self.lc, self.lc_historic = self._update_lora_cache(dry_run=read_from_cache)
        print("Done with LoRa caching: {}".format(t()))

        # Create a set of users with engagements for faster filtering
        engagements = self.lc_historic.engagements.values()
        self.users_with_engagements = set(map(lambda eng: eng[0]["user"], engagements))

        print("Retrieve AD Writer name list")
        with catchtime() as t:
            self.ad_writer = self._get_adwriter(
                lc=self.lc,
                lc_historic=self.lc_historic,
                skip_occupied_names=skip_occupied_names_check,
            )
        print("Done with AD Writer init: {}".format(t()))

        logger.debug("__init__() done")

    def _load_settings(self):
        return load_settings()

    def _load_jinja_filters(self, source: str) -> List[Callable]:
        seeded_create_filters = partial(
            create_filters, tuple_keys=("employee", "ad_object")
        )
        setting_name = f"integrations.ad.lifecycle.{source}"
        filter_templates = self._settings.get(setting_name, [])
        return [
            # Decorate each `filter_func` so it will log skipped users under
            # a name such as "create_filters_num_0", etc.
            self.log_skipped(f"{source}_num_{num}")(filter_func)
            for num, filter_func in enumerate(seeded_create_filters(filter_templates))
        ]

    def _get_adreader(self):
        return ADParameterReader()

    def _get_adwriter(self, **kwargs):
        return ADWriter(**kwargs)

    def log_skipped(self, filtername):
        """Return decorated version of a filter function taking a single
        `tup` arg, which is an `(employee, ad_object)` tuple.
        If the filter function returns `False`, store the result in the
        `stats["skipped"][filtername]` dictionary by the employee UUID.
        """

        def get_employee_name(employee):
            if "name" in employee:
                return " ".join(employee["name"])
            elif "navn" in employee:
                return employee["navn"]
            else:
                return "unknown"

        def decorator(f):
            @wraps(f)
            def wrapper(tup):
                # Call the filter function saving its status
                status = f(tup)
                if status is False:
                    skipped = self.stats.setdefault("skipped", {})
                    users = skipped.setdefault(filtername, {})
                    # Add user UUID to dictionary (name is used for the value)
                    employee = tup[0]
                    users[employee["uuid"]] = get_employee_name(employee)
                return status

            return wrapper

        return decorator

    def _update_lora_cache(self, dry_run: bool = True) -> Tuple[LoraCache, LoraCache]:
        """
        Read all information from AD and LoRa.
        :param dry_run: If True, LoRa dump will be read from cache.
        """
        lc = LoraCache(resolve_dar=False, full_history=False)
        lc.populate_cache(dry_run=dry_run, skip_associations=True)
        lc.calculate_derived_unit_data()
        lc.calculate_primary_engagements()

        lc_historic = LoraCache(resolve_dar=False, full_history=True, skip_past=True)
        lc_historic.populate_cache(dry_run=dry_run, skip_associations=True)

        return lc, lc_historic

    def _gen_stats(self) -> Dict[str, Any]:
        return {
            "critical_errors": 0,
            "engagement_not_found": 0,
            "created_users": 0,
            "disabled_users": 0,
            "already_in_ad": 0,
            "no_active_engagements": 0,
            "not_in_user_tree": 0,
            "create_filtered": 0,
            "users": set(),
        }

    @apply
    def _find_user_unit_tree(self, user: dict, ad_object: dict) -> bool:
        try:
            (
                employment_number,
                title,
                eng_org_unit_uuid,
                eng_uuid,
            ) = self.ad_writer.datasource.find_primary_engagement(user["uuid"])
        except (NoActiveEngagementsException, NoPrimaryEngagementException):
            logger.warning(
                "Warning: Unable to find primary for {}!".format(user["uuid"])
            )
            return False

        logger.debug("Primary found, now find org unit location")

        try:
            unit = self.lc.units[eng_org_unit_uuid][0]
        except KeyError:
            logger.warning(
                "cannot find unit %r (user=%r)", eng_org_unit_uuid, user["uuid"]
            )
            return False

        # Walk up the organisation unit tree, starting at `unit["parent"]`.
        # Stop when we find an allowed root node, or if we encounter a node
        # without a parent (must be root?)
        looking = True
        while looking:
            if unit["uuid"] in self.roots:
                return True
            if unit["parent"] is None:
                return False

            if unit["parent"] in self.lc.units:
                unit = self.lc.units[unit["parent"]][0]
            else:
                logger.warning(
                    "cannot find parent unit %r (user=%r)", unit["parent"], user["uuid"]
                )
                looking = False

        return False

    def _gen_filtered_employees(
        self, in_filters: Optional[List[FilterFunction]] = None
    ):
        def enrich_with_ad_user(mo_employee: dict) -> Tuple[Dict, Dict]:
            """Enrich mo_employee with AD employee dictionary."""
            cpr = mo_employee["cpr"]
            ad_object = self.ad_reader.read_user(cpr=cpr, cache_only=True)
            return mo_employee, ad_object

        @lru_cache(maxsize=0)
        def get_engagements() -> List[LazyDict]:
            """Produce a list of engagements with lazily evaluated properties."""

            def make_class_lazy(class_attribute: str, mo_engagement: dict) -> dict:
                """Create a lazily evaluated class property."""
                class_uuid = mo_engagement[class_attribute]
                mo_engagement[class_attribute + "_uuid"] = class_uuid
                mo_engagement[class_attribute] = LazyEvalBare(
                    lambda: {
                        **self.lc.classes[class_uuid],
                        "uuid": class_uuid,
                    }
                )
                return mo_engagement

            lc_engagements: List[List[Dict]] = self.lc.engagements.values()
            engagements: Iterator[Dict] = map(itemgetter(0), lc_engagements)
            lazy_engagements: Iterator[LazyDict] = map(LazyDict, engagements)
            enriched_engagements: Iterator[LazyDict] = map(
                # Enrich engagement_type class
                partial(make_class_lazy, "engagement_type"),
                map(
                    # Enrich primary_type class
                    partial(make_class_lazy, "primary_type"),
                    map(
                        # Enrich job_function class
                        partial(make_class_lazy, "job_function"),
                        lazy_engagements,
                    ),
                ),
            )
            return list(enriched_engagements)

        def enrich_with_engagements(mo_employee: dict) -> LazyDict:
            """Enrich mo_employee with lazy engagement information.

            The list of engagements is itself lazy, so this code is essentially free
            when it is not in use.
            """
            # Turn mo_employee into a lazy dict and add lazy properties
            lazy_employee: LazyDict = LazyDict(mo_employee)

            lazy_employee["engagements"] = LazyEvalBare(
                lambda: list(
                    filter(
                        lambda engagement: engagement["user"] == mo_employee["uuid"],
                        get_engagements(),
                    )
                )
            )

            lazy_employee["primary_engagement"] = LazyEval(
                lambda key, dictionary: next(
                    filter(
                        lambda engagement: engagement.get("primary_boolean", False),
                        dictionary["engagements"],
                    ),
                    None,
                )
            )

            return lazy_employee

        filters: List[FilterFunction] = in_filters or []

        lc_employees: List[List[Dict]] = self.lc.users.values()
        nonempty_employees = filter(lambda val: len(val) > 0, lc_employees)
        tqdm_employees: List[List[Dict]] = tqdm(nonempty_employees)
        # From employee_effects --> employees
        employees: Iterator[Dict] = map(itemgetter(0), tqdm_employees)

        # Enrich with engagements
        ee_employees: Iterator[Dict] = map(enrich_with_engagements, employees)

        # Enrich with ad_objects
        ad_employees: Iterator[Tuple[Dict, Dict]] = map(
            enrich_with_ad_user, ee_employees
        )

        # Apply requested filters
        for filter_func in filters:
            ad_employees = filter(filter_func, ad_employees)
        return ad_employees

    def disable_ad_accounts(self, dry_run: bool = False) -> Dict[str, Any]:
        """Iterate over all users and disable non-active AD accounts."""

        @apply
        def filter_user_not_in_ad(employee: dict, ad_object: dict) -> bool:
            in_ad = bool(ad_object)
            if not in_ad:
                logger.debug("User {} does not have an AD account".format(employee))
                return False
            return True

        @apply
        def filter_user_has_engagements(employee: dict, ad_object: dict) -> bool:
            # Check the user does not have a valid engagement:
            # TODO: Consider using the lazy properties for this
            if employee["uuid"] in self.users_with_engagements:
                logger.debug("User {} is active - do not touch".format(employee))
                return False
            return True

        employees = self._gen_filtered_employees(
            [
                # Remove users that does not exist in AD
                filter_user_not_in_ad,
                # Remove users that have active engagements
                filter_user_has_engagements,
            ]
            + self.disable_filters
        )
        # Employees now contain only employees which should be disabled
        for employee, ad_object in employees:
            logger.debug("This user has no active engagemens, we should disable")
            # This user has an AD account, but no engagements - disable
            sam = ad_object["SamAccountName"]
            status = True
            message = "dry-run"
            if not dry_run:
                status, message = self.ad_writer.enable_user(username=sam, enable=False)
            if status:
                logger.debug("Disabled: {}".format(sam))
                self.stats["disabled_users"] += 1
                self.stats["users"].add(employee["uuid"])
            else:
                logger.warning("enable_user call failed!")
                logger.warning(message)
                self.stats["critical_errors"] += 1

        return self.stats

    def create_ad_accounts(self, dry_run: bool = False) -> Dict[str, Any]:
        """Iterate over all users and create missing AD accounts."""

        @self.log_skipped("filter_user_already_in_ad")
        @apply
        def filter_user_already_in_ad(employee, ad_object):
            in_ad = bool(ad_object)
            if in_ad:
                self.stats["already_in_ad"] += 1
                logger.debug("User {} is already in AD".format(employee))
                return False
            return True

        @self.log_skipped("filter_user_without_engagements")
        @apply
        def filter_user_without_engagements(employee, ad_object):
            # TODO: Consider using the lazy properties for this
            if employee["uuid"] not in self.users_with_engagements:
                self.stats["no_active_engagements"] += 1
                logger.debug(
                    "User {} has no active engagements - skip".format(employee)
                )
                return False
            return True

        @self.log_skipped("filter_users_outside_unit_tree")
        def filter_users_outside_unit_tree(tup):
            status = self._find_user_unit_tree(tup)
            if status is False:
                self.stats["not_in_user_tree"] += 1
            return status

        def run_create_filters(tup):
            status = all(create_filter(tup) for create_filter in self.create_filters)
            if status is False:
                self.stats["create_filtered"] += 1
            return status

        employees = self._gen_filtered_employees(
            [
                # Remove users that already exist in AD
                filter_user_already_in_ad,
                # Remove users that have no active engagements at all
                filter_user_without_engagements,
                # Check if the user is in a create-user sub-tree
                filter_users_outside_unit_tree,
                # Run all create_filters
                run_create_filters,
            ]
        )
        # Employees now contain only employees which should be created
        for employee, ad_object in employees:
            logger.debug("Create account for {}".format(employee))
            try:
                # Create user without manager to avoid risk of failing
                # if manager is not yet in AD. The manager will be attached
                # by the next round of sync.
                status = True
                message = "dry-run"
                if not dry_run:
                    status, message = self.ad_writer.create_user(
                        employee["uuid"], create_manager=False
                    )
                if status:
                    logger.debug("New username: {}".format(message))
                    self.stats["created_users"] += 1
                    self.stats["users"].add(employee["uuid"])
                else:
                    logger.warning("create_user call failed!")
                    logger.warning(message)
                    self.stats["critical_errors"] += 1
            except NoPrimaryEngagementException:
                logger.exception("No engagment found!")
                self.stats["engagement_not_found"] += 1
            except Exception:
                logger.exception("Unknown error!")
                self.stats["critical_errors"] += 1

        return self.stats


def write_stats(stats: Dict[str, Any]) -> None:
    logger.info("Stats: {}".format(stats))
    stats["users"] = "Written in log file"
    print(stats)


@click.command()
@click.option(
    "--create-ad-accounts",
    default=False,
    is_flag=True,
    help="Create AD Users.",
    type=click.BOOL,
)
@click.option(
    "--disable-ad-accounts",
    default=False,
    is_flag=True,
    help="Disable AD Users.",
    type=click.BOOL,
)
@click.option(
    "--dry-run",
    default=False,
    is_flag=True,
    help="Dry-run without changes.",
    type=click.BOOL,
)
@click.option("--read-from-cache", is_flag=True, envvar="USE_CACHED_LORACACHE")
@click.option(
    "--skip-occupied-names-check",
    default=False,
    is_flag=True,
    help="Skip reading all current user names from AD. Only for testing!",
    type=click.BOOL,
)
def ad_life_cycle(
    create_ad_accounts: bool,
    disable_ad_accounts: bool,
    dry_run: bool,
    read_from_cache: bool,
    skip_occupied_names_check: bool,
) -> None:
    """Create or disable users."""
    logger.debug(
        "Running ad_life_cycle with: {}".format(
            {
                "create_ad_accounts": create_ad_accounts,
                "disable_ad_accounts": disable_ad_accounts,
                "dry_run": dry_run,
                "read_from_cache": read_from_cache,
            }
        )
    )

    if not any([create_ad_accounts, disable_ad_accounts]):
        raise click.ClickException(
            "Either create_ad_accounts or disable_ad_accounts must be given!"
        )

    sync = AdLifeCycle(
        read_from_cache=read_from_cache,
        skip_occupied_names_check=skip_occupied_names_check,
    )

    if create_ad_accounts:
        stats = sync.create_ad_accounts(dry_run)
        write_stats(stats)

    if disable_ad_accounts:
        stats = sync.disable_ad_accounts(dry_run)
        write_stats(stats)


if __name__ == "__main__":
    start_logging("AD_life_cycle.log")
    ad_life_cycle()



import logging
from pathlib import Path

import click
from click_option_group import optgroup
from click_option_group import RequiredMutuallyExclusiveOptionGroup

from .ad_exceptions import NoScriptToExecuteException
from .ad_exceptions import UnknownKeywordsInScriptException
from .ad_writer import ADWriter


logger = logging.getLogger("AdExecute")


EXECUTION_REPLACEMENT_SCAFFOLD = {
    "%OS2MO_AD_BRUGERNAVN%": None,
    "%OS2MO_BRUGER_FORNAVN%": None,
    "%OS2MO_BRUGER_EFTERNAVN%": None,
    "%OS2MO_BRUGER_CPR%": None,
    "%OS2MO_LEDER_EMAIL%": None,
    "%OS2MO_LEDER_NAVN%": None,
    "%OS2MO_BRUGER_ENHED%": None,
    "%OS2MO_BRUGER_ENHED_UUID%": None,
}


class ADExecute(ADWriter):
    def __init__(self):
        super().__init__()
        self.script = None

    def _remove_block_comments(self):
        search_index = 0
        index = self.script.find("<#", search_index)
        remove_ranges = []
        while index > 0:
            end_index = self.script.find("#>", index + 1) + 3
            remove_ranges.append((index, end_index))
            search_index = end_index
            index = self.script.find("<%", search_index)

        for remove_range in remove_ranges:
            self.script = (
                self.script[: remove_range[0]] + self.script[remove_range[1] :]
            )

    def validate_script(self, pre_check=False):
        """
        Validate that the currently loaded script is valid for execution,
        to be valid it must exist, it cannot contain unknown keywords.
        :param pre_check: If true, the errors will be reported back as
        a list, rather than as exceptions.
        """
        if self.script is None:
            msg = "No script loaded!"
            if pre_check:
                return (False, msg)
            logger.error(msg)
            raise NoScriptToExecuteException(msg)

        lines = self.script.split("\n")

        keywords = set()
        for line in lines:
            if line.find("#") > -1:
                line = line[0 : line.find("#")]
            search_index = 0
            index = line.find("%OS2MO_", search_index)
            while index > 0:
                end_index = line.find("%", index + 1)
                keywords.add(line[index : end_index + 1])
                search_index = end_index
                index = line.find("%OS2MO_", search_index)

        keyword_errors = set()
        for keyword in keywords:
            if keyword not in EXECUTION_REPLACEMENT_SCAFFOLD:
                keyword_errors.add(keyword)
        if keyword_errors:
            msg = "Unknown keywords present in template: {}".format(keyword_errors)
            if pre_check:
                return (False, msg)
            logger.error(msg)
            raise UnknownKeywordsInScriptException(msg)
        return (True, "")

    def fill_script_template(self, mo_user_uuid):
        mo_info = self.read_ad_information_from_mo(mo_user_uuid, read_manager=True)
        ad_user = self.get_from_ad(cpr=mo_info["cpr"])
        user_sam = ad_user[0]["SamAccountName"]

        execution_replacement = EXECUTION_REPLACEMENT_SCAFFOLD.copy()
        execution_replacement["%OS2MO_AD_BRUGERNAVN%"] = user_sam
        execution_replacement["%OS2MO_BRUGER_FORNAVN%"] = mo_info["name"][0]
        execution_replacement["%OS2MO_BRUGER_EFTERNAVN%"] = mo_info["name"][1]
        execution_replacement["%OS2MO_BRUGER_CPR%"] = mo_info["cpr"]
        execution_replacement["%OS2MO_LEDER_EMAIL%"] = mo_info["manager_mail"]
        execution_replacement["%OS2MO_LEDER_NAVN%"] = mo_info["manager_name"]
        execution_replacement["%OS2MO_BRUGER_ENHED%"] = mo_info["unit"]
        execution_replacement["%OS2MO_BRUGER_ENHED_UUID%"] = mo_info["unit_uuid"]
        if None in execution_replacement:
            msg = "Not all replacement values available"
            logger.error(msg)
            # TODO: Should we make an explicit exception for this?
            raise Exception(msg)

        actual_script = ""
        lines = self.script.split("\n")
        for line in lines:
            if line.find("#") > -1:
                actual_line = line[0 : line.find("#")]
            else:
                actual_line = line

            for key, replacement in execution_replacement.items():
                actual_line = actual_line.replace(key, replacement)
            if actual_line.strip():
                actual_script += actual_line + "\n"

        self.script = actual_script
        return True

    def read_script_template(self, script_name, pre_check=False):
        if not script_name.endswith("ps_template"):
            script_name += ".ps_template"
        p = Path("scripts/{}".format(script_name))

        if p.is_file():
            self.script = p.read_text()
        validation = self.validate_script(pre_check)
        return validation

    def execute_script(self, script, user_uuid):
        self.read_script_template(script)

        # Remove block comments before performing the validation.
        self._remove_block_comments()

        success = self.fill_script_template(user_uuid)
        if not success:
            msg = "Failed to fill in template"
            logger.error(msg)
            raise Exception(msg)

        response = self._run_ps_script(script)
        if not response == {}:
            msg = "Failed to execute this: {}".format(self.script)
            logger.error(msg)
            msg = "Power Shell error: {}".format(response)
            logger.error(msg)
            raise Exception(msg)
        return "Script completed"


@click.command(help="Powershell Script Executer")
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option(
    "--validate-script",
    help="Validate that a template can be parsed",
)
@optgroup.option(
    "--execute-script",
    help="Execute script with values from user",
    nargs=2,
    type=str,
)
def cli(**args):
    """
    Command line interface for the script executor.
    """
    executor = ADExecute()
    if args.get("validate_script"):
        script = args["validate_script"]
        valid = executor.read_script_template(script, pre_check=True)
        if valid[0]:
            print("Script is valid")
        else:
            print("Script validation failed:\n{}".format(valid[1]))

    if args.get("execute_script"):
        script, user = args["execute_script"]
        print(executor.execute_script(script, user))


if __name__ == "__main__":
    cli()
    # This is a fictious user, Noah Petersen, 111111-1111
    # mo_user = '4931ddb6-5084-45d6-9fb2-52ff33998005'
    # script = 'send_email.ps_template'
    # exe.execute_script(script, mo_user)



import unittest
from datetime import datetime
from datetime import timedelta
from unittest.mock import patch

from parameterized import parameterized

from integrations.rundb.db_overview import DBOverview


class Test_db_overview(unittest.TestCase):
    def setUp(self):
        self.db_overview = DBOverview({"integrations.SD_Lon.import.run_db": "dummy"})

    @parameterized.expand(
        [
            [(datetime.now(), "Lorem Ipsum"), (True, "Status ok")],
            [
                (datetime.now() - timedelta(days=1), "Lorem Ipsum"),
                (False, "Not up to date"),
            ],
            [(datetime.now(), "Running"), (False, "Not ready to run")],
        ]
    )
    @patch("integrations.rundb.db_overview.sqlite3", autospec=True)
    def test_read_current_status(self, fixture_row, expected, sqlite3_mock):
        sqlite3_mock.connect.return_value.cursor.return_value.fetchone.return_value = (
            fixture_row
        )
        status, message = self.db_overview.read_current_status()
        expected_status, expected_message = expected
        self.assertEqual(status, expected_status)
        self.assertEqual(message, expected_message)


if __name__ == "__main__":
    unittest.main()



import csv
from uuid import UUID
from chardet.universaldetector import UniversalDetector


def read_ad_and_uuids():
    file_name = 'AD_Users_Kommune_ALL.csv'

    detector = UniversalDetector()
    with open(file_name, 'rb') as csvfile:
        for row in csvfile:
            detector.feed(row)
            if detector.done:
                break
    detector.close()
    encoding = detector.result['encoding']

    usernames = {}
    uuids = {}
    with open(file_name, encoding=encoding) as csvfile:
        reader = csv.DictReader(csvfile, delimiter=';')
        for row in reader:
            if row['Enabled'] == 'True':
                pass
            else:
                pass
            cpr = row['xAttrCPR']

            usernames[cpr] = row['SamAccountName']
            uuids[cpr] = row['ObjectGUID']
            UUID(uuids[cpr], version=4)  # Fail if not a valid uuid
        
    return usernames, uuids



import json
import pathlib
import sqlite3
from ra_utils.load_settings import load_settings

# TODO:
# This file exists in two quite similar versions.
# Consider to merge into a common tool.


class DBOverview(object):
    def __init__(self):
        self.settings = load_settings()

    def read_db_content(self):
        conn = sqlite3.connect(self.settings['integrations.opus.import.run_db'],
                               detect_types=sqlite3.PARSE_DECLTYPES)
        c = conn.cursor()

        query = 'select * from runs order by id'
        c.execute(query)
        rows = c.fetchall()
        for row in rows:
            status = 'id: {}, dump date: {}, status: {}'
            print(status.format(row[0], row[1], row[2]))

    def read_current_status(self):
        conn = sqlite3.connect(self.settings['integrations.opus.import.run_db'],
                               detect_types=sqlite3.PARSE_DECLTYPES)
        c = conn.cursor()

        query = 'select * from runs order by id desc limit 1'
        c.execute(query)
        row = c.fetchone()
        if 'Running' in row[2]:
            status = (False, 'Not ready to run')
        else:
            status = (True, 'Status ok')
        return status

    def delete_last_row(self, force=False):
        current_status = self.read_current_status()
        if current_status[0] and not force:
            return 'Status ok, no delete'

        conn = sqlite3.connect(self.settings['integrations.opus.import.run_db'],
                               detect_types=sqlite3.PARSE_DECLTYPES)
        c = conn.cursor()
        query = 'select id from runs order by id desc limit 1'
        c.execute(query)
        row = c.fetchone()

        query = 'delete from runs where id=?'
        c.execute(query, (row[0],))
        conn.commit()
        return 'Deleted last row'


if __name__ == '__main__':
    db_overview = DBOverview()

    # print(db_overview.delete_last_row())
    # print(db_overview.delete_last_row(force=True))
    # print(db_overview.delete_last_row(force=True))
    db_overview.read_db_content()
    print(db_overview.read_current_status())



import config
import mox_helpers.payloads as mox_payloads
import uuids
from mox_helpers.mox_helper import create_mox_helper
from mox_helpers.mox_helper import ElementNotFound

from os2mo_data_import import ImportHelper  # type: ignore
from os2mo_data_import.mox_data_types import Itsystem

CLASSES = [
    (
        "Postadresse",
        "org_unit_address_type",
        "DAR",
        "AddressMailUnit",
        uuids.UNIT_POSTADDR,
    ),
    ("LOS ID", "org_unit_address_type", "TEXT", "LOSID", uuids.UNIT_LOS),
    ("CVR nummer", "org_unit_address_type", "TEXT", "CVRUnit", uuids.UNIT_CVR),
    ("EAN nummer", "org_unit_address_type", "EAN", "EANUnit", uuids.UNIT_EAN),
    (
        "P-nummer",
        "org_unit_address_type",
        "PNUMBER",
        "PNumber",
        uuids.UNIT_PNR,
    ),
    ("SE-nummer", "org_unit_address_type", "TEXT", "SENumber", uuids.UNIT_SENR),
    (
        "IntDebitor-Nr",
        "org_unit_address_type",
        "TEXT",
        "intdebit",
        uuids.UNIT_DEBITORNR,
    ),
    ("WWW", "org_unit_address_type", "WWW", "UnitWeb", uuids.UNIT_WWW),
    (
        "Ekspeditionstid",
        "org_unit_address_type",
        "TEXT",
        "UnitHours",
        uuids.UNIT_HOURS,
    ),
    (
        "Telefontid",
        "org_unit_address_type",
        "TEXT",
        "UnitPhoneHours",
        uuids.UNIT_PHONEHOURS,
    ),
    (
        "Telefon",
        "org_unit_address_type",
        "PHONE",
        "UnitPhone",
        uuids.UNIT_PHONE,
    ),
    ("Fax", "org_unit_address_type", "PHONE", "UnitFax", uuids.UNIT_FAX),
    ("Email", "org_unit_address_type", "EMAIL", "UnitEmail", uuids.UNIT_EMAIL),
    (
        "Magkort",
        "org_unit_address_type",
        "TEXT",
        "UnitMagID",
        uuids.UNIT_MAG_ID,
    ),
    (
        "Alternativt navn",
        "org_unit_address_type",
        "TEXT",
        "UnitNameAlt",
        uuids.UNIT_NAME_ALT,
    ),
    (
        "Phone",
        "employee_address_type",
        "PHONE",
        "PhoneEmployee",
        uuids.PERSON_PHONE,
    ),
    (
        "Email",
        "employee_address_type",
        "EMAIL",
        "EmailEmployee",
        uuids.PERSON_EMAIL,
    ),
    (
        "Lokale",
        "employee_address_type",
        "TEXT",
        "RoomEmployee",
        uuids.PERSON_ROOM,
    ),
    ("Primær", "primary_type", "100000", "primary", uuids.PRIMARY),
    ("Ikke-primær", "primary_type", "0", "not_primary", uuids.NOT_PRIMARY),
    (
        "Linjeorganisation",
        "org_unit_hierarchy",
        "TEXT",
        "linjeorg",
        uuids.LINJE_ORG_HIERARCHY,
    ),
    (
        "Sikkerhedsorganisation",
        "org_unit_hierarchy",
        "TEXT",
        "sikkerhedsorg",
        uuids.SIKKERHEDS_ORG_HIERARCHY,
    ),
    (
        "Rolletype",
        "role_type",
        "TEXT",
        "role_type",
        "964c31a2-6267-4388-bff5-42d6f3c5f708",
    ),
    (
        "Tilknytningsrolle",
        "association_type",
        "TEXT",
        "association_type",
        "ec534b86-3d9b-42d8-bff0-afc4f81719af",
    ),
    (
        "Orlovstype",
        "leave_type",
        "TEXT",
        "leave_type",
        "d2892fa6-bc56-4c14-bd24-74ae0c71fa3a",
    ),
    (
        "Alternativ stillingsbetegnelse",
        "employee_address_type",
        "TEXT",
        "AltJobTitle",
        uuids.PERSON_JOB_TITLE_ALT,
    ),
]


async def perform_initial_setup():
    """
    Perform all initial bootstrapping of OS2mo.
    Imports an organisation if missing, and adds all base facets
    Imports all pretedetermined classes and it systems
    """
    settings = config.get_config()
    mox_helper = await create_mox_helper(settings.mox_base)
    try:
        await mox_helper.read_element_organisation_organisation(bvn="%")
    except ElementNotFound:
        print("No org found in LoRa. Performing initial setup.")
        importer = ImportHelper(
            create_defaults=True,
            mox_base=settings.mox_base,
            mora_base=settings.mora_base,
            store_integration_data=False,
            seperate_names=True,
        )
        importer.add_organisation(
            identifier="Århus Kommune",
            user_key="Århus Kommune",
            municipality_code=751,
            uuid=uuids.ORG_UUID,
        )
        # Perform initial import of org and facets
        importer.import_all()

    await import_remaining_classes()
    await import_it()


async def import_remaining_classes():
    """
    Import a set of predetermined classes. All the classes have predefined UUIDs
    which makes this function idempotent
    """
    settings = config.get_config()
    mox_helper = await create_mox_helper(settings.mox_base)

    for clazz in CLASSES:
        titel, facet, scope, bvn, uuid = clazz

        facet_uuid = await mox_helper.read_element_klassifikation_facet(bvn=facet)

        klasse = mox_payloads.lora_klasse(
            bvn=bvn,
            title=titel,
            facet_uuid=str(facet_uuid),
            org_uuid=str(uuids.ORG_UUID),
            scope=scope,
        )
        await mox_helper.insert_klassifikation_klasse(klasse, str(uuid))


async def import_it():
    """
    Import predetermined IT systems. The UUID(s) are predefined which makes this
    function idempotent.
    """
    settings = config.get_config()
    if settings.azid_it_system_uuid == uuids.AZID_SYSTEM:
        mox_helper = await create_mox_helper(settings.mox_base)
        it_system = Itsystem(system_name="AZ", user_key="AZ")
        it_system.organisation_uuid = str(uuids.ORG_UUID)
        uuid = uuids.AZID_SYSTEM
        json = it_system.build()
        await mox_helper.insert_organisation_itsystem(json, str(uuid))
    else:
        print(
            """Settings specify a non-default AZID IT system UUID, not creating
            default AZ IT system"""
        )



import json
import pathlib

import click
from click_option_group import optgroup, RequiredMutuallyExclusiveOptionGroup

from ra_utils.load_settings import load_settings
from integrations.ad_integration.ad_reader import ADParameterReader
from integrations.opus.opus_helpers import update_employee
from integrations.opus.opus_diff_import import start_opus_diff
from integrations.opus.opus_import import start_opus_import
from integrations.opus.opus_exceptions import RunDBInitException
from os2mo_data_import import ImportHelper


@click.command(help="Furesø import")
@optgroup.group("Action", cls=RequiredMutuallyExclusiveOptionGroup)
@optgroup.option("--import", is_flag=True, help='New import into empty MO')
@optgroup.option("--update", is_flag=True, help='Update with next xml file')
@optgroup.option(
    "--update-single-user",
    help='Update a single user, specify employee number',
)
@click.option(
    "--days",
    type=int,
    default=1000,
    help='Number of days in the past to sync single user (default=1000)',
)
def cli(**args):
    SETTINGS = load_settings()

    ad_reader = ADParameterReader()

    if args['update']:
        try:
            start_opus_diff(ad_reader=ad_reader)
        except RunDBInitException:
            print('RunDB not initialized')

    if args['import']:
        importer = ImportHelper(
            create_defaults=True,
            mox_base=SETTINGS['mox.base'],
            mora_base=SETTINGS['mora.base'],
            store_integration_data=False,
            seperate_names=True,
            demand_consistent_uuids=False
        )

        med_name = 'MED Organisation'
        importer.add_klasse(
            identifier=med_name,
            facet_type_ref='org_unit_type',
            user_key=med_name,
            scope='TEXT',
            title=med_name
        )

        importer.add_organisation_unit(
            identifier=med_name,
            name=med_name,
            user_key=med_name,
            type_ref=med_name,
            date_from='1930-01-01',
            date_to=None,
            parent_ref=None
        )

        try:
            start_opus_import(importer, ad_reader=ad_reader, force=True)
        except RunDBInitException:
            print('RunDB not initialized')

    if args['update_single_user']:
        employment_number = args['update_single_user']
        days = args['days']
        update_employee(employment_number, days)


if __name__ == "__main__":
    cli()


import csv
import json
import pathlib
import datetime
from chardet.universaldetector import UniversalDetector

from integrations import cpr_mapper
from os2mo_data_import import ImportHelper
from integrations.SD_Lon import sd_importer
from integrations.ad_integration import ad_reader

# TODO: Soon we have done this 4 times. Should we make a small settings
# importer, that will also handle datatype for specicic keys?
cfg_file = pathlib.Path.cwd() / 'settings' / 'settings.json'
if not cfg_file.is_file():
    raise Exception('No setting file')
settings = json.loads(cfg_file.read_text())

cpr_map = pathlib.Path.cwd() / 'settings' / 'cpr_uuid_map.csv'
if not cpr_map.is_file():
    raise Exception('No mapping file')
employee_mapping = cpr_mapper.employee_mapper(str(cpr_map))

importer = ImportHelper(
    create_defaults=True,
    mox_base=settings['mox.base'],
    mora_base=settings['mora.base'],
    store_integration_data=False,
    seperate_names=True
)

ad_reader = ad_reader.ADParameterReader()

detector = UniversalDetector()
manager_file = settings['integrations.SD_Lon.import.manager_file']
with open(manager_file, 'rb') as csvfile:
    for row in csvfile:
        detector.feed(row)
        if detector.done:
            break
detector.close()
encoding = detector.result['encoding']

manager_rows = []
with open(manager_file, encoding=encoding) as csvfile:
    reader = csv.DictReader(csvfile, delimiter=';')
    for row in reader:
        if row['Leder 1 (cpr-nummer)']:
            new_row = {
                'cpr': row['Leder 1 (cpr-nummer)'].replace('-', ''),
                'ansvar': row['Lederansvar "Leder 1"'],
                'afdeling': row['SD kort navn (afd.kode)']
            }
            manager_rows.append(new_row)

        if row['Leder 2 (cpr-nummer)'].strip():
            new_row = {
                'cpr': row['Leder 2 (cpr-nummer)'].replace('-', ''),
                'ansvar': row['Lederansvar "Leder 2"'],
                'afdeling': row['SD kort navn (afd.kode)']
            }
            manager_rows.append(new_row)


sd = sd_importer.SdImport(
    importer,
    org_only=False,
    ad_info=ad_reader,
    manager_rows=manager_rows,
    employee_mapping=employee_mapping
)

importer.add_klasse(identifier='IT-Org. Alias',
                    uuid='aa4b3520-4ee9-4ac2-9380-a0da852bb538',
                    facet_type_ref='org_unit_address_type',
                    user_key='IT-Org. Alias',
                    scope='TEXT',
                    title='IT-Org. Alias')

sd.create_ou_tree(create_orphan_container=False)
sd.create_employees()

importer.import_all()

"""
for info in sd.address_errors.values():
    print(info['DepartmentName'])
    print(info['DepartmentIdentifier'])
    print(info['PostalAddress']['StandardAddressIdentifier'])
    print(info['PostalAddress']['PostalCode'] + ' ' +
          info['PostalAddress']['DistrictName'])
    print()
    print()
print(len(sd.address_errors))
"""



import pytest
from hypothesis import given
from hypothesis import strategies as st

from ..read_ad_conf_settings import read_settings


def get_minimum_valid_settings():
    return {
        "integrations.ad": [
            {
                "system_user": "system_user",
                "password": "password",
                "search_base": "search_base",
                "cpr_field": "cpr_field",
                "properties": [
                    "l2org_field",
                    "uuid_field",
                    "org_unit_field",
                ],
            }
        ],
        "integrations.ad.winrm_host": "winrm_host",
    }


def get_minimum_valid_writer_settings():
    settings = get_minimum_valid_settings()
    settings.update(
        {
            "integrations.ad.write.level2orgunit_field": "l2org_field",
            "integrations.ad.write.level2orgunit_type": "l2org_type",
            "integrations.ad.write.upn_end": "upn_end",
            "integrations.ad.write.uuid_field": "uuid_field",
            "integrations.ad.write.org_unit_field": "org_unit_field",
        }
    )
    return settings


def test_minimum_valid_settings():
    read_settings(get_minimum_valid_settings())
    read_settings(get_minimum_valid_writer_settings())


def test_missing_winrm_setting():
    settings = get_minimum_valid_settings()
    del settings["integrations.ad.winrm_host"]
    with pytest.raises(Exception) as excinfo:
        read_settings(settings)
        assert "Missing hostname for remote management server" in str(excinfo.value)


@given(st.text())
def test_duplicated_field_names(ad_key):
    settings = get_minimum_valid_writer_settings()
    settings["integrations.ad_writer.mo_to_ad_fields"] = {"a": ad_key}
    settings["integrations.ad_writer.template_to_ad_fields"] = {ad_key: "c"}
    with pytest.raises(ValueError) as excinfo:
        read_settings(settings)
        assert "Duplicated AD field names in settings: [" + ad_key + "]" in str(
            excinfo.value
        )


@given(st.text())
def test_missing_properties(ad_key):
    settings = get_minimum_valid_writer_settings()
    settings["integrations.ad_writer.template_to_ad_fields"] = {ad_key: "c"}
    with pytest.raises(ValueError) as excinfo:
        read_settings(settings)
        assert "Missing AD field names in properties: [" + ad_key + "]" in str(
            excinfo.value
        )
